(*
License: LGPL (see file COPYING.LESSER)
*)
theory Bla
imports Complexity
begin

locale traditional_complexity =
  fixes r :: "('a::size) rel"
begin

abbreviation "DH x \<equiv> {n. \<exists>y. (x, y) \<in> r ^^ n}"

end

locale fb_sn =
  fixes r :: "('a::size) rel"
  assumes fb: "\<And>x. finite {y. (x, y) \<in> r}" and sn: "SN r"
begin

interpretation traditional_complexity r .

lemma finite_DH:
  "finite (DH x)"
using sn
proof (induct)
  case (IH x)
  let ?Y = "\<Union>y \<in> {y. (x, y) \<in> r}. Suc ` DH y"
  have "DH x = {0} \<union> ?Y"
    by (auto simp: relpow_commute elim: relpow_E2 intro!: rev_image_eqI)
       (metis relcomp.relcompI relpow_commute)
  moreover have "finite ?Y"
    using fb and IH by simp
  ultimately show ?case by simp
qed

end

context traditional_complexity
begin

definition "dh x = Max (DH x)"

abbreviation "CP n T \<equiv> {dh x | x. x \<in> T \<and> size x \<le> n}"

definition "cp n T = Max (CP n T)"

abbreviation "of_size A n \<equiv> {x \<in> A. size x \<le> n}"

end

context fb_sn
begin

interpretation traditional_complexity r .

lemma deriv_bound_rel_cp:
  assumes fin: "\<And>n. finite (of_size T n)"
  shows "deriv_bound_rel r (of_size T) (\<lambda>n. cp n T)"
proof (unfold deriv_bound_rel_def, intro allI impI)
  fix x and n
  assume x: "x \<in> of_size T n"
  show "deriv_bound r x (cp n T)"
  proof
    fix y and m
    assume "cp n T < m" and "(x, y) \<in> r ^^ m"
    then have "dh x \<ge> m"
      using finite_DH [of x] by (auto simp: dh_def) (metis Max_ge mem_Collect_eq)
    then have "cp n T \<ge> m"
      using x and finite_image_set [OF fin [of n], of dh]
      by (auto simp: cp_def) (metis (mono_tags, lifting) Max_ge mem_Collect_eq order_trans)
    with \<open>cp n T < m\<close> show False by arith
  qed
qed

lemma cp_lower:
  assumes "\<And>x. DH x \<noteq> {}"
    and "\<And>n. finite (of_size T n)"
    and "\<And>n. CP n T \<noteq> {}"
    and db: "deriv_bound_rel r (of_size T) f"
  shows "f n \<ge> cp n T"
proof (rule ccontr)
  assume "\<not> cp n T \<le> f n"
  then have "cp n T > f n" by arith
  note fin = finite_image_set [OF \<open>finite (of_size T n)\<close>, of dh]
  have *: "\<And>x. x \<in> of_size T n \<Longrightarrow> deriv_bound r x (f n)"
    using db by (auto simp: deriv_bound_rel_def)
  def m \<equiv> "cp n T"
  have **: "\<And>k. k \<in> CP n T \<Longrightarrow> k \<le> m"
    using fin by (auto simp: m_def cp_def)
  have "m \<in> CP n T"
    using Max_in [OF fin \<open>CP n T \<noteq> {}\<close>] by (simp add: m_def cp_def)
  then obtain x where "x \<in> of_size T n" and "dh x = m" by blast
  then have "dh x > f n" by (simp add: m_def) (rule \<open>cp n T > f n\<close>)
  moreover then obtain y where "(x, y) \<in> r ^^ dh x"
    using Max_in [OF finite_DH \<open>DH x \<noteq> {}\<close>] by (auto simp: dh_def)
  ultimately show False using * [OF \<open>x \<in> of_size T n\<close>] by (auto elim: deriv_boundE)
qed

end

definition "foo = {(x, y, z). x = y \<and> y \<le> z}"
definition "bar = {(x, y, z) | x y z. x = y \<and> y \<le> z}"

lemma
  "foo = bar"
  unfolding foo_def bar_def
  apply (simp add: surj_pair)
oops

term "Collect (\<lambda>(x, y, z). x = y \<and> y \<le> z)"
term "{u. \<exists>x y z. u = (x, y, z) \<and> (x = y \<and> y \<le> z)}"
typ "'a \<times> 'b"
term "{f x | x. P x}"
term "wf"
end
