(*
License: LGPL (see file COPYING.LESSER)
*)
theory Nao
imports Peak_Decreasingness
begin

context ars_peak_decreasing
begin

inductive relate where
  "relate [] x x" |
  "(x, y) \<in> r a \<Longrightarrow> relate vs y z \<Longrightarrow> relate (Inr a # vs) x z" |
  "(y, x) \<in> r a \<Longrightarrow> relate vs y z \<Longrightarrow> relate (Inl a # vs) x z"

definition "pm A = (Inl ` A \<union> Inr ` A)"

inductive_set S where
  "(\<And>x y. relate [Inl a, Inr b] x y \<Longrightarrow> relate w x y) \<Longrightarrow> w \<in> lists (pm (\<or>(a, b))) \<Longrightarrow>
    ([Inl a, Inr b], w) \<in> S" |
  "(v, w) \<in> S \<Longrightarrow> (c @ v @ s, c @ w @ s) \<in> S"

fun smash where
  "smash (Inl x) = x" |
  "smash (Inr x) = x"

definition "mof = multiset_of \<circ> map smash"

lemma [iff]:
  "Inl x \<in> pm A \<longleftrightarrow> x \<in> A"
  "Inr x \<in> pm A \<longleftrightarrow> x \<in> A"
  by (auto simp: pm_def)

lemma [dest]:
  assumes "x \<in> pm (\<or>(a, b))"
  shows "a \<succ> smash x \<or> b \<succ> smash x"
  using assms by (cases x) auto

lemma list_pmD [dest]:
  assumes "w \<in> lists (pm (\<or>(a, b)))"
  shows "\<forall>x\<in>set w. a \<succ> smash x \<or> b \<succ> smash x"
  using assms by (induct) auto

lemma mof_mless:
  assumes "\<forall>x\<in>set w. a \<succ> smash x \<or> b \<succ> smash x"
  shows "mof [Inl a, Inr b] \<succ>\<^sub>M mof w"
proof -
  from assms have "\<forall>x. x \<in># mof w \<longrightarrow> a \<succ> x \<or> b \<succ> x"
    by (induct w) (auto simp: mof_def)
  from downset2_multiset [OF this]
    show ?thesis by (simp add: mof_def ac_simps)
qed

lemma S_imp_mof_mless:
  assumes "(w, v) \<in> S"
  shows "mof w \<succ>\<^sub>M mof v"
using assms
apply (induct)
apply (blast dest!: list_pmD mof_mless)
apply (auto simp: mof_def)
by (metis mless_mono union_commute)

lemma SN_S: "SN S"
proof (rule ccontr)
  assume "\<not> SN S"
  then obtain f where "\<forall>i. (f i, f (Suc i)) \<in> S" by auto
  then have "\<forall>i. (mof \<circ> f) i \<succ>\<^sub>M (mof \<circ> f) (Suc i)" by (auto dest: S_imp_mof_mless)
  then have "\<exists>f. \<forall>i. f i \<succ>\<^sub>M f (Suc i)" by blast
  then have "\<not> SN {(x, y). x \<succ>\<^sub>M y}" by auto
  moreover have "SN {(x, y). x \<succ>\<^sub>M y}" by (metis SN_mless)
  ultimately show False by (contradiction)
qed

lemma [simp]:
  "relate [] = (\<lambda>x y. x = y)"
  apply (intro ext)
  by (auto elim: relate.cases intro: relate.intros)

lemma relate_iff:
  "relate (Inl a # ws) x y \<longleftrightarrow> (\<exists>u. (u, x) \<in> r a \<and> relate ws u y)"
  "relate (Inr a # ws) x y \<longleftrightarrow> (\<exists>u. (x, u) \<in> r a \<and> relate ws u y)"
  by (auto elim: relate.cases intro: relate.intros)

lemma relate_append_iff:
  "relate (vs @ ws) x y \<longleftrightarrow> (\<exists>u. relate vs x u \<and> relate ws u y)"
  apply (induct vs arbitrary: x)
  apply simp
  apply (case_tac a)
  by (auto simp: relate_iff)

lemma [simp]:
  "relate (w @ v) x y \<longleftrightarrow> (relate w OO relate v) x y"
  by (auto simp: relate_append_iff)

lemma
  assumes "(u, v) \<in> S"
  shows "\<forall>x y. relate u x y \<longrightarrow> relate v x y"
  using assms by (induct) (auto simp: OO_def)

lemma
  assumes "(x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(x, y) \<in> (\<Union>w\<in>lists (pm I). {(x, y). relate w x y})"
  using assms
  unfolding conversion_def
  by (induct rule: converse_rtrancl_induct)
     (force intro: relate.intros simp: R pm_def)+

end

end

