(*
Author:  Alexander Lochmann <alexander.lochmann@uibk.ac.at> (2018)
License: LGPL (see file COPYING.LESSER)
*)
theory AC_Aux
  imports
    Term_Order
    Signature_Extension
    Multiset_Extension2
    AC_TRS.AC_Equivalence
begin

lemma actop_mset_subterm_eq:
  assumes "x \<in># actop f s"
  shows "s \<unrhd> x"
  using assms
proof (induct f s rule: actop.induct)
  case (1 f g s t)
  then show ?case by (cases "f=g", auto)
qed auto

lemma actop_mset_elem_subterm:
  assumes "x \<in># actop f (Fun f [u,v])"
  shows "Fun f [u,v] \<rhd> x"
  using assms
proof (induct f "Fun f [u,v]" arbitrary: u v rule: actop.induct)
  case (1 f s t)
  then have "x \<in># actop f s \<or> x \<in># actop f t" by simp 
  then show ?case by (metis "1.prems" actop_mset_subterm_eq supt_supteq_conv trivial_Bin_facts(5)) 
qed

lemma rel_subset_trans:
  fixes r :: "nat \<Rightarrow> 'a rel"
  assumes "\<forall> k. r k \<subseteq> r (k+1)"
  and "l \<le> m"
  shows "r l \<subseteq> r m"
  using assms
proof(induct m)
  case 0
  then show ?case by simp
next
  case (Suc m)
  moreover have "r m \<subseteq> r (m+1)" using assms by blast
  ultimately show ?case by (metis Suc_eq_plus1 lift_Suc_mono_le) 
qed

lemma rel_union_trans:
  fixes r :: "nat \<Rightarrow> 'a rel"
  assumes "\<forall> k. trans (r k)"
  and "\<forall> k. r k \<subseteq> r (k+1)"
  shows "trans (\<Union> {r k| k .True})"
proof
  fix x y z
  assume a1:"(x, y) \<in> \<Union>{r k |k. True}"
  assume a2:"(y, z) \<in> \<Union>{r k |k. True}"
  from a1 obtain m where upper_1:"(x,y) \<in> r m" by blast
  from a2 obtain l where upper_2:"(y,z) \<in> r l" by blast
  have "(x,y) \<in> r (max m l) \<and> (y,z) \<in> r (max m l)"
  proof (cases "m > l")
    case True
    then have "r l \<subseteq> r m" using assms(2) rel_subset_trans[of r l m] using less_or_eq_imp_le by blast 
    then show ?thesis using upper_1 upper_2 by (metis True max.strict_order_iff subsetCE)
  next
    case False
    then have "r m \<subseteq> r l" using assms(2) rel_subset_trans[of r m l] not_less by blast 
    then show ?thesis using upper_1 upper_2 by (metis False max_def not_less subsetCE) 
  qed
  then have "(x,z) \<in> r (max m l)" using assms(1) by (meson transE)  
  then show "(x, z) \<in> \<Union>{r k |k. True}" by blast
qed

lemma order_pair_subset_rel:
  fixes s :: "nat \<Rightarrow> 'a rel"
  and  ns :: "'a rel"
  assumes "\<forall> k. order_pair (s k) ns"
  and "\<forall> k. s k \<subseteq> s (k + 1)"
shows "order_pair (\<Union> {s k | k .True}) ns"
proof
  show "refl ns" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  show "trans ns" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  show "trans (\<Union>{s k |k. True})" using assms rel_union_trans[of s] order_pair.axioms(1) pre_order_pair_def by blast 
  show "ns O \<Union>{s k |k. True} \<subseteq> \<Union>{s k |k. True}"
  proof
    fix x y
    assume a:"(x, y) \<in> ns O \<Union>{s k |k. True}"
    then obtain z where con:"(x, z) \<in> ns \<and> (z, y) \<in> \<Union>{s k |k. True}" by blast
    then obtain k where "(z, y) \<in> s k" by blast
    then have "(x, y) \<in> s k" using con assms(1) a order_pair.NS_O_S(1) by blast  
    then show "(x, y) \<in> \<Union>{s k |k. True}" by blast
  qed
  show "\<Union>{s k |k. True} O ns \<subseteq> \<Union>{s k |k. True}"
  proof
    fix x y
    assume a:"(x, y) \<in> \<Union>{s k |k. True} O ns"
    then obtain z where con:"(x, z) \<in> \<Union>{s k |k. True} \<and> (z,y) \<in> ns" by blast
    then obtain k where "(x, z) \<in> s k" by blast
    then have "(x, y) \<in> s k" using con assms(1) order_pair.S_O_NS(1) by blast 
    then show "(x, y) \<in> \<Union>{s k |k. True}" by blast 
  qed
qed


lemma subset_relation:
  shows "U - S \<subseteq># (U - T) + (T - S)"
proof-
  have "U \<subseteq># (U - T) + (T - S) + S"
    by (metis (no_types, lifting) mset_subset_eq_mono_add_left_cancel subset_eq_diff_conv subset_mset.dual_order.refl subset_mset.dual_order.trans union_assoc)
  then show ?thesis using subset_eq_diff_conv by blast
qed

lemma lex_ext_comp_l:
  assumes "order_pair s ns"
  and "snd (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x y)" (is "snd (lex_ext ?f n x y)")
  and "fst (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n y z)"
  shows "fst (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x z)"
proof -
  show ?thesis using assms lex_ext_compat[of x y z ?f n]
    by (smt assms(1) assms(2) assms(3) compat_pair.compat_NS_S_point compat_pair.compat_S_NS_point fst_conv order_pair.axioms(1) order_pair.axioms(2) pre_order_pair.trans_NS_point pre_order_pair.trans_S_point snd_conv) 
qed

lemma lex_ext_comp_r:
  assumes "order_pair s ns"
  and "fst (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x y)" (is "fst (lex_ext ?f n x y)")
  and "snd (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n y z)"
  shows "fst (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x z)"
proof -
  show ?thesis using lex_ext_compat[of x y z ?f n]
    by (smt assms(1) assms(2) assms(3) compat_pair.compat_NS_S_point compat_pair.compat_S_NS_point fst_conv order_pair.axioms(1) order_pair.axioms(2) pre_order_pair.trans_NS_point pre_order_pair.trans_S_point snd_conv) 
qed

lemma lex_ext_s_trans:
  assumes "order_pair s ns"
  and "fst (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x y)" (is "fst (lex_ext ?f n x y)")
  and "fst (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n y z)"
  shows "fst (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x z)"
proof -
  show ?thesis using lex_ext_compat[of x y z ?f n]
    by (smt assms(1) assms(2) assms(3) compat_pair.compat_NS_S_point compat_pair.compat_S_NS_point fst_conv order_pair.axioms(1) order_pair.axioms(2) pre_order_pair.trans_NS_point pre_order_pair.trans_S_point snd_conv) 
qed

lemma lex_ext_ns_trans:
  assumes "order_pair s ns"
  and "snd (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x y)" (is "snd (lex_ext ?f n x y)")
  and "snd (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n y z)"
  shows "snd (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x z)"
proof -
  show ?thesis using lex_ext_compat[of x y z ?f n]
    by (smt assms(1) assms(2) assms(3) compat_pair.compat_NS_S_point compat_pair.compat_S_NS_point fst_conv order_pair.axioms(1) order_pair.axioms(2) pre_order_pair.trans_NS_point pre_order_pair.trans_S_point snd_conv) 
qed

lemma lex_ext_ns_relf:
  assumes "order_pair s ns"
  shows "\<forall> x. snd (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x x)"
proof -
  fix x
  have "snd (lex_ext (\<lambda>l r. ((l, r) \<in> s, (l, r) \<in> ns)) n x x)" using all_nstri_imp_lex_nstri assms order_pair.axioms(1) pre_order_pair.refl_NS_point by fastforce  
  then show ?thesis by (metis (mono_tags, lifting) all_nstri_imp_lex_nstri assms conjI_realizer le_boolD order_pair.axioms(1) order_refl pre_order_pair.refl_NS_point) 
qed

lemma (in order_pair) lex_ext_order_pair:
  shows "order_pair {(x, y). fst (lex_ext (\<lambda> l r. ((l, r) \<in> S, (l, r) \<in> NS)) n x y)}
                    {(x, y). snd (lex_ext (\<lambda> l r. ((l, r) \<in> S, (l, r) \<in> NS)) n x y)}" (is "order_pair ({(x, y). fst (?f n x y)}) ({(x, y). snd (?f n x y)})")
proof
  have ord:"order_pair S NS" by (simp add: order_pair_axioms) 
  show "{(x, y). snd (?f n x y)} O {(x, y). fst (?f n x y)} \<subseteq> {(x, y). fst (?f n x y)}" using lex_ext_comp_l ord by blast
  show "{(x, y). fst (?f n x y)} O {(x, y). snd (?f n x y)} \<subseteq> {(x, y). fst (?f n x y)}" using lex_ext_comp_r ord by blast
  show "trans {(x, y). fst (?f n x y)}" using lex_ext_s_trans[of S NS n] ord by (metis (mono_tags, lifting) case_prodD case_prodI mem_Collect_eq transI) 
  show "trans {(x, y). snd (?f n x y)}" using ord lex_ext_ns_trans[of S NS n] by (metis (mono_tags, lifting) case_prodD case_prodI mem_Collect_eq transI)
  show "refl {(x, y). snd (?f n x y)}" using ord lex_ext_ns_relf[of S NS n] refl_on_def by (simp add: refl_on_def')
qed

lemma lex_ext_s_local_mono:
  assumes "fst (lex_ext (\<lambda>l r. (s l r, ns l r)) n xs ys)" (is "fst (lex_ext ?f _ _ _)")
  and "\<And> i. i < length xs \<Longrightarrow> i < length ys \<Longrightarrow> s (xs ! i) (ys ! i) \<Longrightarrow> s' (xs ! i) (ys ! i)" 
  shows "fst (lex_ext (\<lambda>l r. (s' l r, ns l r)) n xs ys)"
proof -
  let ?f' = "(\<lambda>l r. (s' l r, ns l r))"
  have iff:"(length xs = length ys \<or> length ys \<le> n) \<and> ((\<exists> i < length xs. i < length ys \<and> (\<forall> j < i. snd (?f (xs ! j) (ys ! j))) \<and> fst (?f (xs ! i) (ys !i))) \<or>
  (\<forall> i < length ys. snd (?f (xs ! i) (ys ! i))) \<and> length xs > length ys)" using lex_ext_iff assms(1) by (simp add: lex_ext_iff)  
  then show ?thesis
  proof (cases "\<forall> i < length ys. snd (?f (xs ! i) (ys ! i)) \<and> length xs > length ys")
    case True
    then have "\<forall> i < length ys. snd (?f' (xs ! i) (ys ! i)) \<and> length xs > length ys" by auto
    then show ?thesis using lex_ext_iff[of ?f' n xs ys] using local.iff by auto 
  next
    case False
    then have "(\<exists> i < length xs. i < length ys \<and> (\<forall> j < i. snd (?f (xs ! j) (ys ! j))) \<and> fst (?f (xs ! i) (ys !i)))"
       using False iff assms(1) by blast   
    then obtain i where wit:"i < length xs \<and> i < length ys \<and> (\<forall> j < i. snd (?f (xs ! j) (ys ! j))) \<and> fst (?f (xs ! i) (ys !i))" by blast
    then have "fst (?f' (xs ! i) (ys !i))" using assms(2) by simp
    moreover have "(\<forall> j < i. snd (?f' (xs ! j) (ys ! j)))" using wit by auto 
    ultimately show ?thesis using wit lex_ext_iff[of ?f' n xs ys] iff by auto 
  qed
qed

lemma list_ext_pres_sub_closer:
  assumes "fst (lex_ext (\<lambda> x y. ((x,y) \<in> s_rel, (x, y) \<in> ns_rel)) (length ss) ss ts)"  (is "fst (lex_ext ?f _ _ _)")
  and "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> (s,t) \<in> ns_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ns_rel"
  and "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> (s,t) \<in> s_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> s_rel"
  and "length ss = length ts"
  shows "fst (lex_ext (\<lambda> x y. ((x, y) \<in> s_rel, (x, y) \<in> ns_rel)) (length (map (\<lambda>t. t \<cdot> \<sigma>) ss)) (map (\<lambda>t. t \<cdot> \<sigma>) ss) (map (\<lambda>t. t \<cdot> \<sigma>) ts))"
proof -
  let ?ss' = "map (\<lambda>t. t \<cdot> \<sigma>) ss"
  let ?ts' = "map (\<lambda>t. t \<cdot> \<sigma>) ts"
  have iff:"\<exists> i < length ss. i < length ts \<and> (\<forall> j < i. snd (?f (ss ! j) (ts ! j))) \<and> fst (?f (ss ! i) (ts ! i))"
    using lex_ext_iff assms(1) by (simp add: lex_ext_iff assms(4))
  then have "\<exists> i < length ?ss'. i < length ?ts' \<and> (\<forall> j < i. snd (?f (?ss' ! j) (?ts' ! j))) \<and> fst (?f (?ss' ! i) (?ts' ! i))"
    using assms(2-3) assms(1) leD less_imp_le_nat not_less_eq_eq snd_conv snd_conv by auto
  moreover have "length ?ss' = length ?ts'" using assms(4) by simp
  ultimately show ?thesis unfolding lex_ext_iff fst_conv by blast 
qed

abbreviation sub_mset (infixl "\<cdot>#" 67)
  where
    "sub_mset M s \<equiv> image_mset (\<lambda> x. x \<cdot> s) M"

lemma sub_mset_dist:
  shows "(M + N) \<cdot># s = M \<cdot># s + N \<cdot># s" 
  using image_mset_union by blast

lemma multpw_impl_eq_size:
  assumes "(A, B) \<in> multpw ns_rel"
  shows "size A = size B"
  using assms
proof (induct rule: multpw.induct)
  case empty
  then show ?case by auto
next
  case (add x y X Y)
  then show ?case by auto
qed

lemma ns_s_mul_ext_sum_trans:
  assumes "order_pair s ns"
    and "(A, B + C) \<in> ns_mul_ext ns s"
    and "(B, D) \<in> s_mul_ext ns s"
  shows "(A, D + C) \<in> s_mul_ext ns s"
  using assms
proof -
  from assms(3) have "(B + C, D + C) \<in> s_mul_ext ns s" using assms(1) ns_mul_ext_refl order_pair.axioms(1) pre_order_pair_def s_ns_mul_ext_union_compat by blast
  moreover have "trans s" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast 
  moreover have "trans ns" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  moreover have "compatible_l ns s" by (simp add: assms(1) compatible_l_def order_pair.NS_O_S(1)) 
  moreover have "compatible_r ns s" by (simp add: assms(1) compatible_r_def order_pair.S_O_NS(1)) 
  ultimately show ?thesis using assms ns_s_mul_ext_trans[of s ns A "B + C" "D + C"] using order_pair.axioms(1) pre_order_pair.refl_NS by blast
qed

lemma s_ns_mul_ext_sum_trans:
  assumes "order_pair s ns"
    and "(A, B + C) \<in> s_mul_ext ns s"
    and "(C, D) \<in> ns_mul_ext ns s"
  shows "(A, B + D) \<in> s_mul_ext ns s"
  using assms
proof -
  from assms(3) have s:"(B + C, B + D) \<in> ns_mul_ext ns s" using assms(1) ns_mul_ext_refl ns_ns_mul_ext_union_compat order_pair.axioms(1) pre_order_pair.refl_NS by blast 
  moreover have "trans s" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast 
  moreover have "trans ns" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  moreover have "compatible_l ns s" by (simp add: assms(1) compatible_l_def order_pair.NS_O_S(1)) 
  moreover have "compatible_r ns s" by (simp add: assms(1) compatible_r_def order_pair.S_O_NS(1)) 
  ultimately show ?thesis using assms s  using order_pair.axioms(1) pre_order_pair.refl_NS s_ns_mul_ext_trans by blast 
qed

lemma ns_mul_ext_sum_trans:
  assumes "order_pair s ns"
    and "(A, B + C) \<in> ns_mul_ext ns s"
    and "(B, D) \<in> ns_mul_ext ns s"
  shows "(A, D + C) \<in> ns_mul_ext ns s"
  using assms
proof -
  from assms(3) have "(B + C, D + C) \<in> ns_mul_ext ns s" using assms(1) ns_mul_ext_refl order_pair.axioms(1) pre_order_pair_def ns_ns_mul_ext_union_compat by blast
  moreover have "trans s" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  moreover have "trans ns" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  moreover have "compatible_l ns s" by (simp add: assms(1) compatible_l_def order_pair.NS_O_S(1)) 
  moreover have "compatible_r ns s" by (simp add: assms(1) compatible_r_def order_pair.S_O_NS(1)) 
  ultimately show ?thesis using assms ns_mul_ext_trans[of s ns A "B + C" "D + C"] 
    using order_pair.axioms(1) pre_order_pair.refl_NS by blast
qed

lemma s_mul_ext_sum_trans:
  assumes "order_pair s ns"
    and "(A, B + C) \<in> s_mul_ext ns s"
    and "(B, D) \<in> s_mul_ext ns s"
  shows "(A, D + C) \<in> s_mul_ext ns s"
  using assms
proof -
  from assms(3) have "(B + C, D + C) \<in> ns_mul_ext ns s" using assms(1) by (meson ns_mul_ext_refl ns_ns_mul_ext_union_compat order_pair.axioms(1) pre_order_pair.refl_NS s_ns_mul_ext) 
  moreover have "trans s" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  moreover have "trans ns" using assms(1) order_pair.axioms(1) pre_order_pair_def by blast
  moreover have "compatible_l ns s" by (simp add: assms(1) compatible_l_def order_pair.NS_O_S(1)) 
  moreover have "compatible_r ns s" by (simp add: assms(1) compatible_r_def order_pair.S_O_NS(1)) 
  ultimately show ?thesis using assms s_ns_mul_ext_trans[of s ns A "B + C" "D + C"]
    using order_pair.axioms(1) pre_order_pair.refl_NS by blast
qed

lemma ns_mul_ext_rhs_cancel:
  assumes "order_pair s ns"
  shows "(A + B, A + B - C) \<in> ns_mul_ext ns s"
  using assms by (meson order_pair.axioms(1) pre_order_pair_def subset_eq_diff_conv subset_mset.le_iff_add supseteq_imp_ns_mul_ext)

lemma s_mul_ext_subset_trans:
  assumes "order_pair s ns"
    and "(A, B) \<in> s_mul_ext ns s"
    and "C \<subseteq># B"
  shows "(A, C) \<in> s_mul_ext ns s"
proof -
  have "(B, C) \<in> ns_mul_ext ns s" using assms(1) assms(3) order_pair.axioms(1) pre_order_pair.refl_NS supseteq_imp_ns_mul_ext by blast 
  then show ?thesis by (metis assms(1) assms(2) assms(3) order_pair.axioms(1) order_pair.mul_ext_order_pair pre_order_pair_def subset_mset_def supset_imp_s_mul_ext transD) 
qed

lemma ns_mul_ext_subset_trans:
  assumes "order_pair s ns"
    and "(A, B) \<in> ns_mul_ext ns s"
    and "C \<subseteq># B"
  shows "(A, C) \<in> ns_mul_ext ns s"
proof -
  have "(B, C) \<in> ns_mul_ext ns s" using assms(1) assms(3) order_pair.axioms(1) pre_order_pair.refl_NS supseteq_imp_ns_mul_ext by blast 
  then show ?thesis by (metis assms(1) assms(2) order_pair.axioms(1) order_pair.mul_ext_order_pair pre_order_pair_def transD) 
qed

lemma s_mul_irrefl:
  assumes "ns O s \<subseteq> s"
  and "s O ns \<subseteq> s"
  and irr: "irrefl s"
  and tr: "trans s"
  shows "irrefl (s_mul_ext ns s)"
proof -
  let ?ns = "ns\<^sup>*"
  have sub: "ns \<subseteq> ?ns" by blast
  from assms have comp_r: "?ns O s \<subseteq> s" and comp_l: "s O ?ns \<subseteq> s" apply (simp add: compat_tr_compat)
    by (simp add: assms(1) assms(2) compat_pair.S_O_rtrancl_NS(1) compat_pair.intro) 
  then have "s\<inverse> O ?ns\<inverse> \<subseteq> s\<inverse>" "?ns\<inverse> O s\<inverse> \<subseteq> s\<inverse>"
    by auto
  moreover have "refl ?ns" unfolding refl_on_def by auto
  moreover have "trans ?ns" unfolding trans_def by auto
  moreover have "({#}, {#}) \<notin> mult2_s (?ns\<inverse>) (s\<inverse>)"
    by (simp add: mult2_s_def multpw_converse peak_iff)
  ultimately have "irrefl (s_mul_ext ?ns s)"
    using tr comp_l comp_r irr mult2_s_cancel[of "s\<inverse>" "?ns\<inverse>" "{#}" _ "{#}"]
    by (auto simp: irrefl_def mult2_s_eq_mult2_s_alt s_mul_ext_def)
  then show ?thesis using s_mul_ext_mono[OF sub, of s s] by (meson irrefl_def order_refl subsetCE)
qed

lemma s_mul_ext_pres_sub_closer:
  assumes "(U, V) \<in> s_mul_ext ns_rel s_rel"
    and "\<And> s t. s \<in># U \<Longrightarrow> t \<in># V \<Longrightarrow> (s,t) \<in> ns_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ns_rel"
    and "\<And> s t. s \<in># U \<Longrightarrow> t \<in># V \<Longrightarrow> (s,t) \<in> s_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> s_rel"
  shows "(U \<cdot># \<sigma>,V \<cdot># \<sigma>) \<in> s_mul_ext ns_rel s_rel"
  using assms(1)
proof (rule s_mul_extE)
  let ?f ="\<lambda> x. x \<cdot> \<sigma>"
  fix A1 A2 B1 B2
  assume u:"U = A1 + A2"
  assume v:"V = B1 + B2"
  assume mul_pw:"(A1, B1) \<in> multpw ns_rel"
  assume a:"A2 \<noteq> {#}"
  assume str:"(\<And>b. b \<in># B2 \<Longrightarrow> \<exists>a. a \<in># A2 \<and> (a, b) \<in> s_rel)"
  have str_subs:"\<And> s t. s \<in># A2 \<Longrightarrow> t \<in># B2 \<Longrightarrow> (s,t) \<in> s_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> s_rel" 
    using u v mset_subset_eqD assms(3) by auto
  have ns_subs:"\<And> s t. s \<in># A1 \<Longrightarrow> t \<in># B1 \<Longrightarrow> (s,t) \<in> ns_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ns_rel" 
    using u v mset_subset_eqD assms(2) by auto
  have u_s:"U \<cdot># \<sigma> = A1 \<cdot># \<sigma> + A2 \<cdot># \<sigma>" using u by auto
  moreover have "V \<cdot># \<sigma> = B1 \<cdot># \<sigma> + B2 \<cdot># \<sigma>" using v by auto
  moreover have "(A1 \<cdot># \<sigma>, B1 \<cdot># \<sigma>) \<in> multpw ns_rel" using mul_pw multpw_map[of A1 B1 _ ?f ?f] ns_subs by auto 
  moreover have "A2 \<cdot># \<sigma> \<noteq> {#}" using a by auto
  moreover have "\<And>b. b \<in># B2 \<cdot># \<sigma> \<Longrightarrow> \<exists>a. a \<in># A2 \<cdot># \<sigma> \<and> (a, b) \<in> s_rel" using str_subs str by fastforce 
  ultimately show ?thesis using s_mul_extI by blast
qed


lemma ns_mul_ext_pres_sub_closer:
  assumes "(U, V) \<in> ns_mul_ext ns_rel s_rel"
    and "\<And> s t. s \<in># U \<Longrightarrow> t \<in># V \<Longrightarrow> (s,t) \<in> ns_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ns_rel"
    and "\<And> s t. s \<in># U \<Longrightarrow> t \<in># V \<Longrightarrow> (s,t) \<in> s_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> s_rel"
  shows "(U \<cdot># \<sigma>,V \<cdot># \<sigma>) \<in> ns_mul_ext ns_rel s_rel"
  using assms(1)
proof (rule ns_mul_extE)
  let ?f ="\<lambda> x. x \<cdot> \<sigma>"
  fix A1 A2 B1 B2
  assume u:"U = A1 + A2"
  assume v:"V = B1 + B2"
  assume mul_pw:"(A1, B1) \<in> multpw ns_rel"
  assume str:"(\<And>b. b \<in># B2 \<Longrightarrow> \<exists>a. a \<in># A2 \<and> (a, b) \<in> s_rel)"
  have str_subs:"\<And> s t. s \<in># A2 \<Longrightarrow> t \<in># B2 \<Longrightarrow> (s,t) \<in> s_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> s_rel" 
    using u v mset_subset_eqD assms(3) by auto
  have ns_subs:"\<And> s t. s \<in># A1 \<Longrightarrow> t \<in># B1 \<Longrightarrow> (s,t) \<in> ns_rel \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ns_rel" 
    using u v mset_subset_eqD assms(2) by auto
  have u_s:"U \<cdot># \<sigma> = A1 \<cdot># \<sigma> + A2 \<cdot># \<sigma>" using u by auto
  moreover have "V \<cdot># \<sigma> = B1 \<cdot># \<sigma> + B2 \<cdot># \<sigma>" using v by auto
  moreover have "(A1 \<cdot># \<sigma>, B1 \<cdot># \<sigma>) \<in> multpw ns_rel" using mul_pw multpw_map[of A1 B1 _ ?f ?f] ns_subs by auto 
  moreover have "\<And>b. b \<in># B2 \<cdot># \<sigma> \<Longrightarrow> \<exists>a. a \<in># A2 \<cdot># \<sigma> \<and> (a, b) \<in> s_rel" using str_subs str by fastforce 
  ultimately show ?thesis using ns_mul_extI by blast
qed

lemma vars_term_ms_simp[simp]:
  fixes s :: "('f,'v) term"
  shows "vars_term_ms (Fun f [s]) = vars_term_ms s"
  by (induct s) auto

lemma vars_term_not_empty_imp_no_constant:
  assumes "vars_term_ms (Fun f ss) \<noteq> {#}"
  shows "length ss > 0"
  using assms
  by (cases "length ss = 0") auto

lemma actop_impl_subterm:
  assumes "x \<in># actop f s"
  shows "actop f s = {#s#} \<or> s \<rhd> x"
  using assms
proof (induct s)
  case (Var x)
  then show ?case by auto
next
  case (Fun g ss)
  then show ?case using actop_non_Bin[of ss f g]
    apply auto
    by (metis \<open>\<forall>u v. ss \<noteq> [u, v] \<Longrightarrow> actop f (Fun g ss) = {#Fun g ss#}\<close> actop.simps(1) actop_mset_elem_subterm)
qed

fun vars_ctxt_ms :: "('f,'v)ctxt \<Rightarrow> 'v multiset"
  where
    "vars_ctxt_ms Hole = {#}" |
    "vars_ctxt_ms (More f ss C ts) =
      \<Union># (mset (map vars_term_ms ss)) +
      vars_ctxt_ms C +
      \<Union># (mset (map vars_term_ms ts))"

lemma vars_ctxt_ms:
  shows "vars_term_ms C\<langle>s\<rangle> = vars_ctxt_ms C + vars_term_ms s"
  by (induct C) auto

lemma vars_ms_closed_under_ctxt:
  fixes s :: "('f,'v) term"
  assumes "vars_term_ms s = vars_term_ms t"
  shows "vars_term_ms C\<langle>s\<rangle> = vars_term_ms C\<langle>t\<rangle>"
  using assms vars_ctxt_ms[of C s] vars_ctxt_ms[of C t] by (auto)

lemma vars_ms_closed_under_sub:
  assumes "vars_term_ms s = vars_term_ms t"
  shows "vars_term_ms (s \<cdot> \<sigma>) = vars_term_ms (t \<cdot> \<sigma>)"
  using assms vars_term_ms_subst_mono[of s t \<sigma>] vars_term_ms_subst_mono[of t s \<sigma>] by auto

lemma ac_terms_eq_root:
  assumes "(s, t) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
  shows   "root s = root t"
  using AC_class_abstract_aocnf.aocnf_roots aocnf_abstract_aocnf.acconv_imp_nf_eq assms by blast

lemma ac_rule_share_vars:
  assumes "(l, r) \<in> AC_rules AC AC"
  shows "vars_term_ms l = vars_term_ms r"
  using assms
  by (auto simp: A_rules_def C_rules_def AC_rules_def)

lemma ac_step_share_vars:
  assumes "(s, t) \<in> (acstep AC AC)"
  shows "vars_term_ms s = vars_term_ms t"
  using assms
proof (cases)
  case (rstep C \<sigma> l r)
  then show ?thesis using ac_rule_share_vars[of l r]
    by (auto simp: vars_term_ms_subst_mono vars_ms_closed_under_ctxt vars_ms_closed_under_sub ctxt_closed_acstep_symcl)
qed

lemma ac_converse_share_vars:
  assumes "(s, t) \<in> (acstep AC AC)\<inverse>"
  shows "vars_term_ms s = vars_term_ms t"
  using assms
proof 
  have "(t, s) \<in> acstep AC AC" using assms converseD[of s t] by auto
  then show ?thesis using assms ac_step_share_vars[of t s] by (auto)
qed

lemma ac_symcl_share_vars:
  fixes s ::"('f, 'v) term"
  assumes "(s, t) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>"
  shows "vars_term_ms s = vars_term_ms t"
  using assms ac_step_share_vars ac_converse_share_vars by auto

lemma ac_terms_share_vars:
  assumes "(s, t) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "vars_term_ms s = vars_term_ms t"
  using assms
  unfolding conversion_def by induction (auto simp:ac_symcl_share_vars)

lemma acstep_closed_subst:
  fixes s ::"('f, 'v) term"
    and \<sigma> ::"'v \<Rightarrow> ('f, 'v) term"
  assumes "(s,t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
proof -
  have "subst.closed ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    unfolding conversion_def using symcl_acstep subst_closed_rsteps[of "AC_trs AC AC"] by auto
  then show ?thesis using assms subst by (simp add: \<open>subst.closed ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)\<close> subst.closedD)
qed

lemma ac_vars_eq:
  assumes "(s, t) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "is_Var s"
    and "is_Var t"
  shows "s = t"
  using assms
proof -
  obtain x where term_s:"s = Var x" using assms(2) by blast 
  obtain y where term_t:"t = Var y" using assms(3) by blast 
  have "vars_term_ms s = vars_term_ms t" using assms(1) ac_terms_share_vars[of s t] by blast
  thus ?thesis by (simp add: term_s term_t)
qed

lemma not_ac_imples_not_aceq_actop_ms:
  assumes "(s, t) \<notin> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "f \<in> AC"
  shows "image_mset (acnf AC) (actop f s) \<noteq> image_mset (acnf AC) (actop f t)"
  using assms
proof (induct s)
  case (Var x)
  then show ?case by (meson AC_class_abstract_acnf.acnf_eq_intro acconv_iff)
next
  case (Fun x1a x2)
  then show ?case by (meson AC_class_abstract_acnf.acnf_eq_intro acconv_iff) 
qed

lemma ac_no_ac_root_on_args_list:
  assumes "(Fun f ss, Fun f ts) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "f \<notin> AC"
  shows "map (acnf AC) ss = map (acnf AC) ts"
proof -
  have eq:"acnf AC (Fun f ss) = acnf AC (Fun f ts)" by (meson acnf_abstract_acnf.acconv_imp_nf_eq assms(1))
  thus ?thesis
  proof(cases "(Fun f ss)" rule:acnf.cases)
    case (1 x)
    then show ?thesis using eq by blast 
  next
    case (2 g u v)
    then have w:"acnf AC (Fun f ss) = ABin g (acnf AC u) (acnf AC v)" using eq assms(2) by auto
    moreover have "\<exists> x y. acnf AC (Fun f ts) = ABin g (acnf AC x) (acnf AC y)" using eq assms(2) 2 by (metis calculation)
    then obtain x y where w2:"acnf AC (Fun f ts) = ABin g (acnf AC x) (acnf AC y)" by blast
    moreover have "acnf AC u = acnf AC x" using w w2 eq by simp
    moreover have "acnf AC v = acnf AC y" using w w2 eq by simp
    ultimately show ?thesis using eq assms(2) 
      by (smt acnf.simps(3) acnf_code(2) acterm.inject(2) list.simps(8) list.simps(9) mset_map) 
  next
    case (3 f ts)
    then show ?thesis using eq by (smt AC_class_abstract_aocnf.aocnf_not_ABin acnf.simps(3) acterm.inject(2) aocnf_abstract_aocnf.acconv_imp_nf_eq aocnf_code(2) assms(1) assms(2) mset_map term.inject(2)) 
  qed
qed

lemma ac_arity_not_two_impl_acnf_eq_on_args_list:
  assumes "(Fun f ss, Fun f ts) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "length ss \<noteq> 2"
  shows "map (acnf AC) ss = map (acnf AC) ts"
proof -
  have eq:"acnf AC (Fun f ss) = acnf AC (Fun f ts)" by (meson acnf_abstract_acnf.acconv_imp_nf_eq assms(1))
  thus ?thesis
  proof(cases "(Fun f ss)" rule:acnf.cases)
    case (1 x)
    then show ?thesis using eq by blast 
  next
    case (2 g u v)
    show ?thesis using eq assms(2) 2 by auto 
  next
    case (3 f ts)
    then show ?thesis using eq
      by (smt One_nat_def Suc_eq_plus1 ac_terms_eq_root acnf.simps(3) acterm.inject(2) assms(1) assms(2) list.size(3) list.size(4) nat_1_add_1 root.simps(2) root_Some term.inject(2)) 
  qed
qed

lemma ac_arity_not_two_impl_ac_terms_on_args_list:
  assumes "(Fun f ss, Fun f ts) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "length ss \<noteq> 2"
  shows "i < length ss \<Longrightarrow> (ss ! i, ts ! i) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
proof
  assume a:"i < length ss"
  have eq:"map (acnf AC) ss = map (acnf AC) ts" using ac_arity_not_two_impl_acnf_eq_on_args_list assms by metis 
  then show "(ss ! i, ts ! i) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>)\<^sup>*" by (meson a acconv_iff conversionE map_nth_conv) 
qed

lemma ac_rule_arg_of_unary_fun:
  assumes "(Fun s [s1] , Fun t [t1]) \<in> AC_rules AC AC"
  shows "(s1, t1) \<in> AC_rules AC AC"
  using assms
  by (auto simp: A_rules_def C_rules_def AC_rules_def)

lemma acstep_arg_of_unary_fun:
  assumes "(Fun f [s1], Fun f [t1]) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(s1, t1) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  have "acnf AC (Fun f [s1]) = AFun f [acnf AC s1]" by simp 
  moreover have "acnf AC (Fun f [t1]) = AFun f [acnf AC t1]" by simp
  moreover have "acnf AC (Fun f [s1]) = acnf AC (Fun f [t1])" using assms(1) using acconv_iff by blast
  ultimately have "acnf AC s1 = acnf AC t1" by simp
  thus ?thesis by (simp add: acconv_iff) 
qed

definition filter_fun :: "('f, 'v) Term.term multiset \<Rightarrow> (('f \<times> nat) \<Rightarrow> ('f \<times> nat) \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat) \<Rightarrow> ('f, 'v) Term.term multiset"
  where
    "filter_fun T P f = filter_mset (\<lambda> x. (case x of (Var _) \<Rightarrow> False | (Fun g ts) \<Rightarrow> P (g, length ts) f)) T"

(*abbreviation of filtering *)
abbreviation filter_var :: "('f, 'v) term multiset \<Rightarrow> ('f, 'v) term multiset"
  where
    "filter_var T \<equiv> filter_mset (\<lambda> x. is_Var x) T"

notation filter_var ("_ \<restriction>\<^sub>v" [200] 100)

lemma filter_fun_add_mset_order:
  assumes "P (g, length gs) f"
    and "s = Fun g gs" 
  shows "filter_fun (add_mset s S) P f = add_mset s (filter_fun S P f)"
  using assms unfolding filter_fun_def by (induct S) auto

lemma filter_fun_add_mset_cancel:
  assumes "\<not> P (g, length gs) f"
    and "s = Fun g gs" 
  shows "filter_fun (add_mset s S) P f = filter_fun S P f"
  using assms unfolding filter_fun_def by (induct S) auto

lemma filter_fun_var_cancel:
  assumes "is_Var s" 
  shows "filter_fun (add_mset s S) P f = filter_fun S P f"
  using assms unfolding filter_fun_def by (induct S) auto

lemma filter_fun_impl_subeq:
  shows "filter_fun S P f \<subseteq># S"
  unfolding filter_fun_def by auto

lemma filter_fun_dist:
  shows "filter_fun (A + B) P f = filter_fun A P f + filter_fun B P f"
  unfolding filter_fun_def by simp

lemma filter_var_of_fun_empty:
  assumes "\<And> x. x \<in># S \<and> is_Fun x"
  shows "S\<restriction>\<^sub>v = {#}"
  using assms by simp

lemma filter_fun_of_var_empty:
  assumes "\<And> x. x \<in># T \<and> is_Var x"
  shows "filter_fun T P f = {#}"
  using assms unfolding filter_fun_def by blast

lemma filter_fun_distj_filter_var:
  shows "S\<restriction>\<^sub>v \<inter># filter_fun T P f = {#}"
  unfolding filter_fun_def by (simp add: disjunct_not_in term.case_eq_if) 

lemma filter_minus_assoc:
  shows "filter_fun T P f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v = filter_fun T P f + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)"
proof -
  have "filter_fun T P f \<inter># S\<restriction>\<^sub>v = {#}" by (simp add: filter_fun_distj_filter_var multiset_inter_commute)
  then show ?thesis by (simp add: multiset_union_diff_commute union_commute)
qed

locale prec =
  fixes pr_strict :: "('f \<times> nat) \<Rightarrow> ('f \<times> nat) \<Rightarrow> bool"
begin

(*
  Note that we use the following orientation order
  x > y \<longleftrightarrow> pr_strict x y
  x < y \<longleftrightarrow> pr_strict y x
*)
abbreviation filter_func_less :: "('f, 'v) term multiset \<Rightarrow> 'f \<times> nat \<Rightarrow>('f, 'v) term multiset"
  where
    "filter_func_less T f \<equiv> filter_fun T (\<lambda> x y. pr_strict y x) f"

abbreviation filter_func_nless :: "('f, 'v) term multiset \<Rightarrow> 'f \<times> nat \<Rightarrow>('f, 'v) term multiset"
  where
    "filter_func_nless T f \<equiv> filter_fun T (\<lambda> x y. \<not> pr_strict y x) f"

notation filter_func_less ("_ \<restriction>\<^sub>p _" [200, 200] 100)
notation filter_func_nless ("_ \<restriction>\<^sub>n _" [200, 200] 100)

abbreviation filtered_less_rel_s :: "('f, 'v) term multiset \<Rightarrow> ('f, 'v) term multiset \<Rightarrow> 'f \<times> nat \<Rightarrow> ('f, 'v) term rel \<Rightarrow> ('f, 'v) term rel \<Rightarrow> bool"
  where
    "filtered_less_rel_s S T f ns_ord s_ord \<equiv> (S\<restriction>\<^sub>p f, T\<restriction>\<^sub>p f) \<in> s_mul_ext ns_ord s_ord"

abbreviation filtered_nless_rel_s ::  "('f, 'v) term multiset \<Rightarrow> ('f, 'v) term multiset \<Rightarrow> 'f \<times> nat \<Rightarrow> ('f, 'v) term rel \<Rightarrow> ('f, 'v) term rel \<Rightarrow> bool"
  where
    "filtered_nless_rel_s  S T f ns_ord s_ord \<equiv> (S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> s_mul_ext ns_ord s_ord"

abbreviation filtered_nless_rel_ns :: "('f, 'v) term multiset \<Rightarrow> ('f, 'v) term multiset \<Rightarrow> 'f \<times> nat \<Rightarrow> ('f, 'v) term rel \<Rightarrow> ('f, 'v) term rel \<Rightarrow> bool"
  where
    "filtered_nless_rel_ns S T f ns_ord s_ord \<equiv> (S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns_ord s_ord"

abbreviation ackbo_case3_filtered_rel
  where
    "ackbo_case3_filtered_rel S T f ns_rel s_rel \<equiv> filtered_nless_rel_s S T f ns_rel s_rel \<or>
           filtered_nless_rel_ns S T f ns_rel s_rel \<and> size S > size T \<or>
           filtered_nless_rel_ns S T f ns_rel s_rel \<and> size S = size T \<and>
           filtered_less_rel_s S T f ns_rel s_rel"

abbreviation ackbo_case3_filtered
  where
    "ackbo_case3_filtered S T f ns_rel s \<equiv>
        let (sr,ns) = mulextp (\<lambda> t u.(s t u, (t,u) \<in> ns_rel)) (S\<restriction>\<^sub>n f) (T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) in
           if sr then True else
           if ns \<and> size S > size T then True else
           if ns \<and> size S = size T
           then smulextp (\<lambda> t u.(s t u, (t,u) \<in> ns_rel)) (S\<restriction>\<^sub>p f) (T\<restriction>\<^sub>p f)
           else False"

lemma filter_unions_eq:
  shows "S = S\<restriction>\<^sub>v + S\<restriction>\<^sub>n f + S\<restriction>\<^sub>p f"
proof(induct S)
  case empty
  then show ?case unfolding filter_fun_def by simp
next
  case (add x S)
  then consider (a) "is_Var x"| (b) "\<exists> g hs. x = Fun g hs \<and> pr_strict f (g,length hs)"|
    (c) "\<exists> g hs. x = Fun g hs \<and> \<not> (pr_strict f (g,length hs))" by (meson is_FunE)
  then show ?case
  proof cases
    case a
    then have "add_mset x S \<restriction>\<^sub>n f + add_mset x S \<restriction>\<^sub>p f = S \<restriction>\<^sub>n f + S \<restriction>\<^sub>p f" unfolding filter_fun_def by (simp add: term.case_eq_if)
    then show ?thesis using add by (metis a filter_mset_add_mset union_assoc union_mset_add_mset_left) 
  next
    case b
    then have "add_mset x S \<restriction>\<^sub>p f = S \<restriction>\<^sub>p f + {#x#}" unfolding filter_fun_def by auto
    moreover have "add_mset x S \<restriction>\<^sub>n f + add_mset x S \<restriction>\<^sub>v = S \<restriction>\<^sub>n f + S \<restriction>\<^sub>v" unfolding filter_fun_def using b by auto 
    ultimately show ?thesis using add.hyps b by auto
  next
    case c
    then have "add_mset x S \<restriction>\<^sub>n f = S \<restriction>\<^sub>n f + {#x#}" unfolding filter_fun_def by auto
    moreover have "add_mset x S \<restriction>\<^sub>p f + add_mset x S \<restriction>\<^sub>v = S \<restriction>\<^sub>p f + S \<restriction>\<^sub>v" unfolding filter_fun_def using c by auto 
    ultimately show ?thesis using add.hyps c by auto
  qed
qed


lemma filtered_s_imp_filter_ns:
  assumes "filtered_nless_rel_s S T f ns s "
  shows "filtered_nless_rel_ns S T f ns s"
  using assms s_ns_mul_ext by blast

lemma filter_compact_l:
  assumes "order_pair s ns"
    and "filtered_nless_rel_ns S T f ns s "
    and "filtered_nless_rel_s T U f ns s"
  shows "filtered_nless_rel_s S U f ns s"
proof -
  from assms(1) have ref:"refl ns" using order_pair.axioms(1) pre_order_pair.refl_NS by blast
  then have "locally_refl ns (T\<restriction>\<^sub>v)" using r_lr by blast
  have S:"(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)) \<in> ns_mul_ext ns s" using assms(2) by (simp add: filter_minus_assoc) 
  have T:"(T\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + (U\<restriction>\<^sub>v - T\<restriction>\<^sub>v)) \<in> s_mul_ext ns s" using assms(3)
    using filter_minus_assoc by (simp add: assms(3) filter_minus_assoc)
  then have "((U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v), U\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns s" 
    using subset_relation supseteq_imp_ns_mul_ext[of ns "U\<restriction>\<^sub>v - S\<restriction>\<^sub>v" "(U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)" s] ref by auto
  moreover have "(S\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + (U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)) \<in> s_mul_ext ns s" using assms S T ns_s_mul_ext_sum_trans[of s ns] by blast
  ultimately have "(S\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + U\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> s_mul_ext ns s" using assms(1) s_ns_mul_ext_sum_trans[of s ns "S\<restriction>\<^sub>n f" "U\<restriction>\<^sub>n f"] by (metis filter_minus_assoc union_assoc)
  then show ?thesis by auto
qed

lemma filter_compact_r:
  assumes "order_pair s ns"
    and "filtered_nless_rel_s S T f ns s "
    and "filtered_nless_rel_ns T U f ns s"
  shows "filtered_nless_rel_s S U f ns s"
proof -
  have S:"(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)) \<in> s_mul_ext ns s" using assms(2) by (simp add: filter_minus_assoc) 
  have T:"(T\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + (U\<restriction>\<^sub>v - T\<restriction>\<^sub>v)) \<in> ns_mul_ext ns s" using assms(3) using filter_minus_assoc by (simp add: assms(3) filter_minus_assoc)
  then have "((U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v), U\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns s" using subset_relation supseteq_imp_ns_mul_ext[of ns "U\<restriction>\<^sub>v - S\<restriction>\<^sub>v" "(U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)" s] assms(1) order_pair.axioms(1) pre_order_pair_def by blast 
  moreover have "(S\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + (U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)) \<in> s_mul_ext ns s" using assms S T s_ns_mul_ext_sum_trans[of s ns]  by (metis (no_types, lifting) union_commute) 
  ultimately have "(S\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + U\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> s_mul_ext ns s" using assms(1) s_ns_mul_ext_sum_trans[of s ns "S\<restriction>\<^sub>n f" "U\<restriction>\<^sub>n f"] by (metis filter_minus_assoc union_assoc)
  then show ?thesis by blast 
qed

lemma filter_trans_ns:
  assumes "order_pair s ns"
    and "filtered_nless_rel_ns S T f ns s"
    and "filtered_nless_rel_ns T U f ns s"
  shows "filtered_nless_rel_ns S U f ns s"
proof -
  have S:"(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)) \<in> ns_mul_ext ns s" using assms(2) by (simp add: filter_minus_assoc) 
  have T:"(T\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + (U\<restriction>\<^sub>v - T\<restriction>\<^sub>v)) \<in> ns_mul_ext ns s" using assms(3) using filter_minus_assoc by (simp add: assms(3) filter_minus_assoc)
  then have "((U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v), U\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns s" using subset_relation supseteq_imp_ns_mul_ext[of ns "U\<restriction>\<^sub>v - S\<restriction>\<^sub>v" "(U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)" s] assms(1) order_pair.axioms(1) pre_order_pair_def by blast 
  moreover have "(S\<restriction>\<^sub>n f, U\<restriction>\<^sub>n f + (U\<restriction>\<^sub>v - T\<restriction>\<^sub>v) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)) \<in> ns_mul_ext ns s" using assms S T ns_mul_ext_sum_trans[of s ns] by blast
  ultimately have "(S\<restriction>\<^sub>n f,U\<restriction>\<^sub>v - S\<restriction>\<^sub>v + U\<restriction>\<^sub>n f ) \<in> ns_mul_ext ns s" using assms(1) ns_mul_ext_sum_trans[of s ns "S\<restriction>\<^sub>n f" "U\<restriction>\<^sub>v - T\<restriction>\<^sub>v + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)" "U\<restriction>\<^sub>n f" "U\<restriction>\<^sub>v - S\<restriction>\<^sub>v"]
    by (simp add: add.left_commute union_commute) 
  then show ?thesis by (metis filter_minus_assoc union_commute)
qed

lemma filter_trans_s:
  assumes "order_pair s ns"
    and "filtered_nless_rel_s S T f ns s"
    and "filtered_nless_rel_s T U f ns s"
  shows "filtered_nless_rel_s S U f ns s"
proof -
  show ?thesis using assms filter_compact_r[OF assms(1) assms(2)] s_ns_mul_ext by blast
qed

lemma A2:
  assumes "order_pair s ns"
  shows "order_pair {(S, T). filtered_nless_rel_s S T f ns s} {(S, T). filtered_nless_rel_ns S T f ns s}" (is "order_pair ?S_Rel ?NS_Rel")
proof
  have ord_pair_mul:"order_pair (s_mul_ext ns s) (ns_mul_ext ns s)" using assms order_pair.mul_ext_order_pair[of s ns] by auto
  then have pre_ord_pair_mul:"pre_order_pair (s_mul_ext ns s) (ns_mul_ext ns s)" using order_pair.axioms[of "(s_mul_ext ns s)" "(ns_mul_ext ns s)"] by auto
  show "refl ?NS_Rel" using pre_ord_pair_mul pre_order_pair.refl_NS_point refl_on_def by fastforce
  show "?NS_Rel O ?S_Rel \<subseteq> ?S_Rel" using filter_compact_l using assms by blast
  show "?S_Rel O ?NS_Rel \<subseteq> ?S_Rel" using filter_compact_r assms by blast
  show "trans ?S_Rel" unfolding trans_def using filter_trans_s assms by blast
  show "trans ?NS_Rel" unfolding trans_def using filter_trans_ns assms by blast
qed

lemma exist_elem_right_ns:
  assumes "(L, R) \<in> multpw ns"
    and "e \<in># L"
    and "\<forall> l \<in># R. (e, l) \<notin> ns"
  shows "False"
proof -
  have "L \<noteq> {#}" using assms(2) by auto 
  then have "(\<exists>x y X Y. L = add_mset x X \<and> R = add_mset y Y \<and> (x, y) \<in> ns \<and> (X, Y) \<in> multpw ns)" using assms(1) multpw.simps[of L R ns] by blast 
  then show "False" using assms by (metis mset_add multpw_split1R union_single_eq_member)
qed

lemma multpw_impl_ns_mul_ext:
  assumes "(S, T) \<in> multpw ns"
  shows "(S, T) \<in> ns_mul_ext ns str"
  by (simp add: assms multpw_converse multpw_implies_mult2_alt_ns ns_mul_ext_def)

lemma acstep_on_var_impl_var_eq:
  assumes "(s,t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "s = Var x"
  shows "t = Var x"
  using assms by (metis ac_terms_eq_root ac_vars_eq is_FunE root.simps(2) root_Some term.distinct(1)) 

lemma image_mset_multpw_cong:
  assumes "image_mset f S = image_mset f T"
  shows "(S,T) \<in> multpw {(x,y). f x = f y}"
  using assms 
proof(induct S arbitrary: T)
  case empty
  then show ?case by simp 
next
  case (add x S)
  then show ?case using multpw.simps[of "add_mset x S" T "{(x,y). f x = f y}"]
    by (smt case_prodI image_mset_add_mset mem_Collect_eq msed_map_invR)
qed

lemma multpw_impl_eq_size:
  assumes "(S,T) \<in> multpw ns"
  shows "size S = size T"
  using assms by (induct rule:multpw.induct) auto

lemma multpw_acstep_impl_filter_var_eq:
  assumes "(S, T) \<in> multpw ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "S\<restriction>\<^sub>v = T\<restriction>\<^sub>v"
  using assms
proof(induct rule:multpw.induct)
  case empty
  then show ?case by auto
next
  case (add x y X Y)
  then show ?case using acstep_on_var_impl_var_eq
    by (metis converse.intros conversion_converse filter_mset_add_mset is_Var_def) 
qed


lemma acstep_impl_actop_multpw:
  assumes "(s,t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "f \<in> AC"
  shows "(actop f s, actop f t) \<in> multpw ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
proof -
  have "image_mset (acnf AC) (actop f s) = image_mset (acnf AC) (actop f t)"
    by (metis acnf_abstract_acnf.acconv_imp_nf_eq assms(1) assms(2) image_mset_acnf_actop)
  then have "(actop f s, actop f t) \<in> multpw {(x, y). (acnf AC) x = (acnf AC) y}" using image_mset_multpw_cong by blast  
  then show ?thesis using acconv_iff by (smt Collect_cong Collect_mem_eq case_prodE case_prodI2) 
qed

lemma multpw_acstep_impl_filter_fun_multpw:
  assumes "(S, T) \<in> multpw ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "(filter_fun S R f,filter_fun T R f) \<in> multpw ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  using assms
proof(induct rule:multpw.induct)
  case empty
  then show ?case unfolding filter_fun_def by auto
next
  case (add x y X Y)
  have eq:"root x = root y" using ac_terms_eq_root add(1) by blast
  then show ?case
  proof (cases "is_Fun x")
    case True
    then have "\<exists> g gs. x = Fun g gs" by blast
    then obtain g gs where term_x:"x = Fun g gs" by auto
    have "\<exists> ts. y = Fun g ts" using eq term_x by (metis prod.sel(1) root.simps(2) root_Some)   
    then obtain ts where term_y:"y = Fun g ts" by auto
    have len:"length ts = length gs" using eq term_x term_y by auto
    consider (a) "R (g,length gs) f" | (b) "\<not> R (g,length gs) f" by auto
    then show ?thesis
    proof cases
      case a
      have "filter_fun (add_mset x X) R f = add_mset x (filter_fun X R f)"  using term_x filter_fun_add_mset_order[of R, OF a] by blast
      moreover have "filter_fun (add_mset y Y) R f = add_mset y (filter_fun Y R f)" using term_y len a filter_fun_add_mset_order by metis 
      ultimately show ?thesis using add(3) by (simp add: add.hyps(1) multpw.add) 
    next
      case b
      then have "filter_fun (add_mset x X) R f = filter_fun X R f" using term_x filter_fun_add_mset_cancel[of R, OF b] by blast
      moreover have "filter_fun (add_mset y Y) R f = filter_fun Y R f" using term_y len b filter_fun_add_mset_cancel by metis
      ultimately show ?thesis using add by auto
    qed 
  next
    case False
    then have "is_Var x \<and> is_Var y" using eq term.distinct is_Fun_Fun_conv by fastforce 
    then show ?thesis using filter_fun_var_cancel add by metis
  qed
qed

lemma acstep_impl_filtered_nless_rel_ns:
  assumes "(s,t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "s = Fun f ss"
    and "t = Fun f ts"
    and "length ss = length ts"
    and "f \<in> AC"
  shows "filtered_nless_rel_ns (actop f s) (actop f t) (f, length ss) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) str"
proof -
  let ?S = "actop f s"
  let ?T = "actop f t"
  let ?f = "(f, length ss)"
  have act:"(?S, ?T) \<in> multpw ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" using acstep_impl_actop_multpw assms by (simp add: acstep_impl_actop_multpw) 
  then have "?S\<restriction>\<^sub>v = ?T\<restriction>\<^sub>v" using multpw_acstep_impl_filter_var_eq by blast 
  moreover have "(?S\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f) \<in> multpw ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" using act by (simp add: act multpw_acstep_impl_filter_fun_multpw assms(4)) 
  ultimately have "(?S\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f + ?T\<restriction>\<^sub>v - ?S\<restriction>\<^sub>v) \<in> multpw ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" by auto
  then have "(?S\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f + ?T\<restriction>\<^sub>v - ?S\<restriction>\<^sub>v) \<in> ns_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) str" using multpw_impl_ns_mul_ext by blast
  then show ?thesis by (simp)
qed

lemma s_mul_ext_local_mono:
  fixes S ::"('f, 'v) term multiset"
  assumes "(S,T) \<in> s_mul_ext ns str"
    and "\<And> s t.  s \<in># S \<Longrightarrow> t \<in># T \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
  shows "(S, T) \<in> s_mul_ext ns str'" using assms
  by (auto elim!: s_mul_extE) (meson s_mul_extI)

lemma ns_mul_ext_local_mono:
  fixes S ::"('f, 'v) term multiset"
  assumes "(S,T) \<in> ns_mul_ext ns str"
    and "\<And> s t.  s \<in># S \<Longrightarrow> t \<in># T \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
  shows "(S, T) \<in> ns_mul_ext ns str'" using assms
  by (auto elim!: ns_mul_extE) (meson ns_mul_extI)

lemma distj_msets_msum_presv_subeq:
  assumes "A \<inter># B = {#}"
    and "A \<subseteq># C"
    and "B \<subseteq># C"
  shows "A + B \<subseteq># C"
  using assms by (metis diff_intersect_left_idem diff_zero subset_eq_diff_conv subset_mset.diff_add subset_mset.le_diff_conv2)

lemma subset_rel_filterd_comb:
  shows "T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v \<subseteq># T"
proof -
  have "T\<restriction>\<^sub>n f + T \<restriction>\<^sub>v - S\<restriction>\<^sub>v = T\<restriction>\<^sub>n f + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)" using filter_minus_assoc by blast 
  moreover have "T\<restriction>\<^sub>n f \<subseteq># T" using filter_fun_impl_subeq by blast
  moreover have "T \<restriction>\<^sub>v - S\<restriction>\<^sub>v \<subseteq># T" using multiset_filter_subset diff_subset_eq_self subset_mset.dual_order.trans by blast
  ultimately show ?thesis using filter_fun_distj_filter_var distj_msets_msum_presv_subeq[of "T\<restriction>\<^sub>n f" "T\<restriction>\<^sub>v - S\<restriction>\<^sub>v"]
    by (metis filter_diff_mset multiset_inter_commute) 
qed

lemma filtered_nless_rel_s_local_mono:
  fixes S ::"('f, 'v) term multiset"
  assumes "filtered_nless_rel_s S T f ns str"
    and "\<And> s t.  s \<in># S \<Longrightarrow> t \<in># T \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
  shows "filtered_nless_rel_s S T f ns str'"
proof -
  have "S\<restriction>\<^sub>n f \<subseteq># S \<and> (T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<subseteq># T"
    using filter_fun_impl_subeq multiset_filter_subset subset_rel_filterd_comb by blast
  then have ord:"\<And> s t. s \<in># S\<restriction>\<^sub>n f \<Longrightarrow> t \<in># T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
    using assms(2) by (meson mset_subset_eqD)
  have w:"(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> s_mul_ext ns str" using assms by (simp)
  then have "(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> s_mul_ext ns str'" using s_mul_ext_local_mono[OF w] ord by blast
  then show ?thesis by auto
qed

lemma filtered_nless_rel_ns_local_mono:
  fixes S ::"('f, 'v) term multiset"
  assumes "filtered_nless_rel_ns S T f ns str "
    and "\<And> s t.  s \<in># S \<Longrightarrow> t \<in># T \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
  shows "filtered_nless_rel_ns S T f ns str'"
proof -
  have "S\<restriction>\<^sub>n f \<subseteq># S \<and> (T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<subseteq># T"
    using filter_fun_impl_subeq multiset_filter_subset subset_rel_filterd_comb by blast
  then have ord:"\<And> s t. s \<in># S\<restriction>\<^sub>n f \<Longrightarrow> t \<in># T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
    using assms(2) by (meson mset_subset_eqD)
  have "(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns str" using assms by (simp)
  from ns_mul_ext_local_mono[OF this] have "(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns str'" using ord by auto
  then show ?thesis by simp 
qed

lemma filtered_less_rel_s_local_mono:
  fixes S ::"('f, 'v) term multiset"
  assumes "filtered_less_rel_s S T f ns str"
    and "\<And> s t.  s \<in># S \<Longrightarrow> t \<in># T \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
  shows "filtered_less_rel_s S T f ns str'"
proof -
  have "S\<restriction>\<^sub>n f \<subseteq># S \<and> (T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<subseteq># T"
    using filter_fun_impl_subeq multiset_filter_subset subset_rel_filterd_comb by blast
  then have ord:"\<And> s t. s \<in># S\<restriction>\<^sub>n f \<Longrightarrow> t \<in># T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v \<Longrightarrow> (s, t) \<in> str \<Longrightarrow> (s, t) \<in> str'"
    using assms(2) by (meson mset_subset_eqD)
  have w:"(S\<restriction>\<^sub>p f,T\<restriction>\<^sub>p f) \<in> s_mul_ext ns str" using assms by (simp)
  have "(S\<restriction>\<^sub>p f,T\<restriction>\<^sub>p f) \<in> s_mul_ext ns str'" using s_mul_ext_local_mono[OF w] ord
    by (meson assms(2) filter_fun_impl_subeq mset_subset_eqD) 
  then show ?thesis by simp
qed

lemma filtered_less_rel_s_cond_trans:
  assumes "order_pair s ns"
    and "filtered_less_rel_s S T f ns s"
    and "filtered_less_rel_s T U f ns s"
  shows "filtered_less_rel_s S U f ns s"
proof -
  have S:"(S\<restriction>\<^sub>p f, T\<restriction>\<^sub>p f) \<in> s_mul_ext ns s" using assms(2) by (simp) 
  moreover have T:"(T\<restriction>\<^sub>p f, U\<restriction>\<^sub>p f) \<in> s_mul_ext ns s" using assms(3) by (simp)
  ultimately have "(S\<restriction>\<^sub>p f, U\<restriction>\<^sub>p f) \<in> s_mul_ext ns s" using assms(1) by (metis (no_types, lifting) order_pair.axioms(1) order_pair.mul_ext_order_pair pre_order_pair_def transD) 
  then show ?thesis by auto
qed

lemma acstep_filter_r_compact_r:
  assumes "order_pair str ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "(s,t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "f \<in> AC"
    and "s = Fun f ss"
    and "filtered_less_rel_s (actop f t) (actop f u) (f,length ss) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) str" (is "filtered_less_rel_s ?T ?U ?f ?ns _")
  shows "filtered_less_rel_s (actop f s) (actop f u) (f,length ss) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) str"
proof -
  let ?S = "actop f s"
  have "(?S\<restriction>\<^sub>p ?f, ?T\<restriction>\<^sub>p ?f) \<in> ns_mul_ext ?ns str"
    using assms(2) acstep_impl_actop_multpw by (simp add: acstep_impl_actop_multpw assms(3) multpw_acstep_impl_filter_fun_multpw multpw_impl_ns_mul_ext) 
  moreover have "(?T\<restriction>\<^sub>p ?f, ?U\<restriction>\<^sub>p ?f) \<in> s_mul_ext ?ns str"
    using assms(5) by blast
  ultimately have "(?S\<restriction>\<^sub>p ?f,?U\<restriction>\<^sub>p ?f) \<in> s_mul_ext ?ns str" 
    using ns_s_mul_ext_sum_trans[OF assms(1),of _ _ "{#}"] by auto
  then show ?thesis by blast
qed

lemma acstep_filter_r_compact_l:
  assumes "order_pair str ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "f \<in> AC"
    and "s = Fun f ss"
    and "filtered_less_rel_s (actop f s) (actop f t) (f,length ss) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) str" (is "filtered_less_rel_s ?S ?T ?f ?ns _")
    and "(t,u) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "filtered_less_rel_s (actop f s) (actop f u) (f,length ss) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) str"
proof -
  let ?U = "actop f u"
  have "(?T\<restriction>\<^sub>p ?f, ?U\<restriction>\<^sub>p ?f)\<in> ns_mul_ext ?ns str"
    using assms(5) acstep_impl_actop_multpw by (simp add: acstep_impl_actop_multpw assms(2) multpw_acstep_impl_filter_fun_multpw multpw_impl_ns_mul_ext) 
  moreover have "(?S\<restriction>\<^sub>p ?f, ?T\<restriction>\<^sub>p ?f) \<in> s_mul_ext ?ns str" using assms(4) by blast
  ultimately have "(?S\<restriction>\<^sub>p ?f, ?U\<restriction>\<^sub>p ?f) \<in> s_mul_ext ?ns str" 
    using s_ns_mul_ext_sum_trans[OF assms(1),of _ "{#}"] by auto
  then show ?thesis by simp
qed

lemma ackbo_case3_filtered_rel_irrefl:
  assumes "ns_rel O s_rel \<subseteq> s_rel"
  and "s_rel O ns_rel \<subseteq> s_rel"
  and irr: "irrefl s_rel"
  and tr: "trans s_rel"
shows "irrefl {(S,T).ackbo_case3_filtered_rel S T f ns_rel s_rel}"
proof -
  from s_mul_irrefl[OF assms] have "irrefl (s_mul_ext ns_rel s_rel)" by simp
  then show ?thesis unfolding irrefl_def by auto
qed

lemma ackbo_case3_closed_under_mset_union:
  assumes "ackbo_case3_filtered_rel S T f ns_rel s_rel "
    and "order_pair s_rel ns_rel"
  shows "ackbo_case3_filtered_rel (S + U) (T + U) f ns_rel s_rel"
proof -
  have sub:"T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v \<subseteq># T" using subset_rel_filterd_comb by (simp add: filter_minus_assoc)
  have U_dist:"(T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) + U\<restriction>\<^sub>n f = (T\<restriction>\<^sub>n f + U\<restriction>\<^sub>n f) + (T\<restriction>\<^sub>v - S\<restriction>\<^sub>v)"
    using filter_minus_assoc union_assoc union_commute
    by smt 
  have Var_dist:"(T + U)\<restriction>\<^sub>v - (S + U)\<restriction>\<^sub>v = T\<restriction>\<^sub>v - S\<restriction>\<^sub>v" using filter_union_mset add_diff_cancel_right by auto
  consider (a) "filtered_nless_rel_s S T f ns_rel s_rel "|
    (b) "filtered_nless_rel_ns S T f ns_rel s_rel  \<and> size T < size S"|
    (c) "filtered_nless_rel_ns S T f ns_rel s_rel  \<and> size S = size T \<and>
                  filtered_less_rel_s S T f ns_rel s_rel"
    using assms by blast
  then show ?thesis
  proof cases
    case a
    then have "(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> s_mul_ext ns_rel s_rel" by auto
    then have "(S\<restriction>\<^sub>n f + U\<restriction>\<^sub>n f, (T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) + U\<restriction>\<^sub>n f) \<in> s_mul_ext ns_rel s_rel"
      using s_mul_ext_union_compat[of "S\<restriction>\<^sub>n f" "T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v" ns_rel s_rel "U\<restriction>\<^sub>n f"] assms(2) pre_order_pair_def r_lr  order_pair.axioms(1) by blast 
    then have "filtered_nless_rel_s (S + U) (T + U) f ns_rel s_rel" using U_dist Var_dist by (metis filter_fun_dist filter_minus_assoc)
    then show ?thesis by auto
  next
    case b
    then have "(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns_rel s_rel" by auto
    then have "(S\<restriction>\<^sub>n f + U\<restriction>\<^sub>n f, (T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) + U\<restriction>\<^sub>n f) \<in> ns_mul_ext ns_rel s_rel"
      using ns_mul_ext_union_compat[of "S\<restriction>\<^sub>n f" "T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v" ns_rel s_rel "U\<restriction>\<^sub>n f"] assms(2) pre_order_pair_def r_lr  order_pair.axioms(1) by blast 
    then have "filtered_nless_rel_ns (S + U) (T + U) f ns_rel s_rel" using U_dist Var_dist by (metis filter_fun_dist filter_minus_assoc)
    moreover have "size (S + U) > size (T + U)" using b by auto
    ultimately show ?thesis by auto
  next
    case c
    then have "(S\<restriction>\<^sub>p f, T\<restriction>\<^sub>p f) \<in> s_mul_ext ns_rel s_rel" by auto 
    then have r_vers:"(S\<restriction>\<^sub>p f + U\<restriction>\<^sub>p f, T\<restriction>\<^sub>p f + U\<restriction>\<^sub>p f) \<in> s_mul_ext ns_rel s_rel"
      using s_mul_ext_union_compat assms(2) pre_order_pair_def r_lr order_pair.axioms(1) by blast
    have "(S\<restriction>\<^sub>n f, T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<in> ns_mul_ext ns_rel s_rel" using c  by auto
    then have "(S\<restriction>\<^sub>n f + U\<restriction>\<^sub>n f, (T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) + U\<restriction>\<^sub>n f) \<in> ns_mul_ext ns_rel s_rel"
      using ns_mul_ext_union_compat[of "S\<restriction>\<^sub>n f" "T\<restriction>\<^sub>n f + T\<restriction>\<^sub>v - S\<restriction>\<^sub>v" ns_rel s_rel "U\<restriction>\<^sub>n f"] assms(2) pre_order_pair_def r_lr order_pair.axioms(1) by blast 
    then have "filtered_nless_rel_ns (S + U) (T + U) f ns_rel s_rel " using U_dist Var_dist by (metis filter_fun_dist filter_minus_assoc)
    moreover have "size (S + U) = size (T + U)" using c by auto 
    ultimately show ?thesis using r_vers by (simp add: filter_fun_dist) 
  qed
qed

lemma filtered_nless_rel_s:
  assumes "order_pair s_rel ns_rel"
    and "\<forall>y. y \<in># M \<longrightarrow> SN_on s_rel {y}"
  shows "SN_on {(S,T). filtered_nless_rel_s S T h ns_rel s_rel} {M}"
proof (rule ccontr)
  let ?cond = "\<lambda> S T. filtered_nless_rel_s S T h ns_rel s_rel"
  assume a:"\<not> SN_on {(S,T). ?cond  S T} {M}"
  then obtain f where f: "f 0 = M" "\<And> n :: nat. ?cond (f n) (f (Suc n))" unfolding SN_defs by auto
  then have "\<And> n ::nat. (f n\<restriction>\<^sub>n h + f n\<restriction>\<^sub>v,(f (Suc n)\<restriction>\<^sub>n h + f (Suc n)\<restriction>\<^sub>v - f n\<restriction>\<^sub>v) + f n\<restriction>\<^sub>v) \<in> s_mul_ext ns_rel s_rel" 
    by (metis add_cancel_left_left assms(1) diff_union_cancelL mset_subset_eq_exists_conv ns_mul_ext_rhs_cancel s_ns_mul_ext_union_compat) 
  then have sn:"\<And> n ::nat. (f n\<restriction>\<^sub>n h + f n\<restriction>\<^sub>v,f (Suc n)\<restriction>\<^sub>n h + f (Suc n)\<restriction>\<^sub>v) \<in> s_mul_ext ns_rel s_rel"
    using s_mul_ext_subset_trans[OF assms(1)]
    by (smt assms(1) diff_zero s_mul_ext_subset_trans subset_relation)
  let ?f = "\<lambda> x. f x\<restriction>\<^sub>n h + f x\<restriction>\<^sub>v" 
  have "\<forall> y. y \<in># M\<restriction>\<^sub>n h + M\<restriction>\<^sub>v \<longrightarrow> SN_on s_rel {y}" by (metis assms(2) filter_unions_eq union_iff)
  from SN_s_mul_ext_strong[OF assms(1) this] have "SN_on (s_mul_ext ns_rel s_rel) {M\<restriction>\<^sub>n h + M\<restriction>\<^sub>v}" by auto
  then show False
    using sn chain_imp_not_SN_on[of ?f "s_mul_ext ns_rel s_rel" 0] f
    by (auto)
qed

end

end
