(*
Author:  Alexander Lochmann <alexander.lochmann@uibk.ac.at> (2018)
License: LGPL (see file COPYING.LESSER)
*)
theory Ackbo_simp_ord
  imports
    Ackbo_split
    Simple_Termination
begin

context admissible_weight_fun_ac
begin

lemma min_weigth_of_not_empty_list:
  assumes "ss \<noteq> []"
  shows "sum_list (map weight ss) \<ge> w0"
proof (cases ss)
  case Nil
  then show ?thesis using assms(1) by auto
next
  case (Cons a ss)
  then show ?thesis by (simp add: trans_le_add1 weight_w0) 
qed

lemma min_weigth_of_list_ge_1:
  assumes "u \<in> set ss"
    and "length ss > 1"
  shows "sum_list (map weight ss) \<ge> weight u + w0"
proof (cases ss)
  case Nil
  then show ?thesis using assms(1) by auto
next
  case (Cons a ss)
  then show ?thesis
  proof (cases "u \<in> set ss")
    case True
    then have "sum_list (map weight ss) \<ge> weight u" by (induct ss) auto
    then have "sum_list (map weight (a #ss)) \<ge> weight u + w0" by (simp add: weight_w0 add.commute add_le_mono) 
    then show ?thesis using Cons by auto
  next
    case False
    then have "\<exists> x. x \<in> set ss" using Cons assms(2) by (metis One_nat_def less_add_same_cancel2 list.size(4) nth_mem)
    then have s:"sum_list (map weight ss) \<ge> w0" using Cons by (metis list.distinct(1) list.set_cases min_weigth_of_not_empty_list)
    from False have "u = a" using assms(1) Cons by fastforce
    then have "sum_list (map weight (a #ss)) \<ge> weight u + w0" using s by (simp add: weight_w0 add.commute add_le_mono) 
    then show ?thesis using Cons by auto
  qed
qed

lemma subterm_impl_fun:
  assumes "s \<rhd> t"
  shows "is_Fun s"
  using assms by blast

lemma subterm_weigth_eq_impl_unary_fun:
  assumes "s \<rhd> t"
    and "weight s = weight t"
  shows "length (args s) = 1"
proof (rule ccontr)
  assume a:"num_args s \<noteq> 1"
  obtain f ss where term_s:"s = Fun f ss" using assms(1) subterm_impl_fun by blast
  have "\<exists> u \<in> set ss. u \<unrhd> t" using assms(1) \<open>s = Fun f ss\<close> by blast
  then obtain u where f:"u \<in> set ss \<and> u \<unrhd> t" by blast
  then have "num_args s > 1" using a length_removeAll_less term_s by fastforce
  then have "sum_list (map weight ss) \<ge> weight u + w0" using term_s a weight_w0 f min_weigth_of_list_ge_1 by auto
  moreover have "weight u = weight t" using f assms(2) term_s  by (metis le_antisym subterm_weight supt.arg weight_subterm)
  ultimately show "False" using assms(2) term_s w0(2) by auto 
qed

lemma ackbo_closed_under_lhs_ext:
  shows "ackbo (Fun f [s]) s"
proof(induct s)
  case (Var x)
  then show ?case
    by (metis ackbo_case0.intros(1) ackbo_case0_lhs_contains_var case_0 case_w nat_less_le weight.simps(1) weight_w0) 
next
  case (Fun g hs)
  consider (a) "w(f,1) = 0" | (b) "w(f,1) \<noteq> 0" by blast
  then show ?case using Fun
  proof cases
    case a
    then show ?thesis
    proof (cases "f = g \<and> length hs = 1")
      case True
      then obtain h1 where sub:"hs = [h1]" by (metis One_nat_def Suc_length_conv length_0_conv) 
      then have "ackbo (Fun g hs) h1" using Fun True by auto
      then show ?thesis using True sub ackbo_trivial_context_closure by blast 
    next
      case False
      then have "pr_strict (f,1) (g,length hs)" using adm a by auto
      have w:"weight (Fun f [Fun g hs]) = weight (Fun g hs)" using a by auto
      have "vars_term_ms (Fun g hs) \<subseteq># vars_term_ms (Fun f [Fun g hs])" by auto
      from case_1[OF this w] show ?thesis apply auto
        using One_nat_def \<open>pr_strict (f, 1) (g, length hs)\<close> by presburger 
    qed
  next
    case b
    then have "weight (Fun f [s]) > weight s" by auto
    then show ?thesis by (simp add: case_w) 
  qed
qed

lemma ackbo_supt_one:
  assumes "ackbo s t"
  shows "ackbo (Fun f (bef @ s # aft)) t"
proof -
  let ?args = "bef @ s # aft"
  let ?lhs = "Fun f ?args"
  have vars:"vars_term_ms s \<subseteq># vars_term_ms ?lhs" by simp
  have "ackbo ?lhs s"
  proof -
    consider (a) "weight ?lhs > weight s" | (b) "weight ?lhs = weight s" by fastforce
    then show ?thesis
    proof cases
      case a
      then show ?thesis using ackbo.case_w vars by fastforce
    next
      case b
      have "weight (Fun f (bef @ s # aft)) = w(f, length (bef @ s # aft)) + weight s + sum_list (map weight bef) + sum_list (map weight aft)" by simp
      then have empty:"bef = [] \<and> aft = []" using b
        by (metis add_cancel_left_right leD le_add1 le_add2 le_antisym min_weigth_of_not_empty_list trans_le_add1 w0(2)) 
      then have wf:"w(f, 1) = 0" using b by simp
      then have "weight (Fun f [s]) = weight s" using b by auto
      then have "ackbo (Fun f [s]) s"
      proof(induct s)
        case (Var x)
        then show ?case using b vars empty by (simp add: case_0 ackbo_case0.intros(1))
      next
        case (Fun g hs)
        then show ?case
        proof (cases "f = g \<and> length hs = 1")
          case True
          then obtain h1 where sub:"hs = [h1]" by (metis One_nat_def Suc_length_conv length_0_conv) 
          then have "ackbo (Fun g hs) h1" using Fun True by auto
          then show ?thesis using True sub ackbo_trivial_context_closure by blast 
        next
          case False
          then have "pr_strict (f,1) (g,length hs)" using adm wf by auto
          then show ?thesis using Fun(2) by (simp add: case_1) 
        qed
      qed
      then show ?thesis using empty by auto
    qed
  qed
  moreover have "trans {(x,y). ackbo x y}" using ackbo_ord_p.order_pair_axioms order_pair_def pre_order_pair_def by auto
  ultimately show ?thesis using assms(1) unfolding trans_def by blast
qed

lemma ackbo_subterm_property:
  assumes "s \<rhd> t"
  shows "ackbo s t"
proof -
  obtain C where s: "s = C\<langle>t\<rangle>" and C: "C \<noteq> \<box>" using assms by auto
  show ?thesis unfolding s using C
  proof (induct C arbitrary: t)
    case Hole thus ?case by simp
  next
    case (More f bef C aft t)
    show ?case
    proof (cases "C = \<box>")
      case True
      consider (a) "bef = [] \<and> aft = []" | (b) "\<not> (bef = [] \<and> aft = [])" by blast
      then show ?thesis
      proof cases
        case a
        then show ?thesis using ackbo_closed_under_lhs_ext unfolding True by fastforce
      next
        case b
        then have "0 < sum_list (map weight bef) \<or> 0 < sum_list (map weight aft)"
          by (metis gr_zeroI leD min_weigth_of_not_empty_list w0(2)) 
        moreover have "weight (Fun f (bef @ t # aft)) = w(f, length (bef @ t # aft)) + weight t + sum_list (map weight bef) + sum_list (map weight aft)" by simp
        ultimately have "weight (Fun f (bef @ t # aft)) > weight t" by auto
        then show ?thesis unfolding True using case_w
          by (simp add: \<open>\<And>t s. \<lbrakk>vars_term_ms t \<subseteq># vars_term_ms s; weight t < weight s\<rbrakk> \<Longrightarrow> ackbo s t\<close>) 
      qed
    next
      case False
      then show ?thesis using More ackbo_supt_one by fastforce
    qed
  qed
qed

lemma actop_mim_size:
  shows "size (actop f t) \<ge> 1"
proof (induct t)
  case (Var x)
  then show ?case by auto
next
  case (Fun f ss)
  then show ?case using actop.simps(1)
    by (metis (no_types, lifting) actop_singleton eq_iff list.set_intros(1) size_single size_union term.inject(2) trans_le_add1) 
qed

lemma actop_size_one:
  assumes "size (actop f t) = 1"
  shows "actop f t = {#t#}"
  using assms apply (cases rule:actop.cases) apply (auto)
  by (metis Suc_n_not_le_n assms actop.simps(1) actop_mim_size actop_singleton non_empty_plus_non_empty_not_single size_1_singleton_mset size_eq_0_iff_empty) 

lemma aux_for_ctxt_closure:
  assumes "(s,t) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "f \<in> AC"
  shows "({#s#}, actop f t) \<in> ns_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) {(x,y). ackbo x y} \<or> ({#s#}, actop f t) \<in> s_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) {(x,y). ackbo x y}"
proof (cases "size (actop f t) = 1")
  case True
  then have "actop f t = {#t#}" using actop_size_one by auto
  then show ?thesis by (simp add: assms(1)) 
next
  case False
  then have sm:"\<And>b. b \<in># actop f t \<Longrightarrow> (\<exists>a. a \<in># {#t#} \<and> ackbo a b)" using ackbo_subterm_property actop_impl_subterm
    by (metis multi_member_last size_single) 
  then have "({#t#}, actop f t) \<in> s_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) {(x,y). ackbo x y} " 
    using s_mul_extI[of "{#t#}" "{#}" "{#t#}" "actop f t" "{#}" "actop f t" "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" "{(x, y). ackbo x y}"] by auto 
  then show ?thesis using sm by (metis  ackbo_ord_p.compat_NS_S_point add_mset_eq_single add_mset_not_empty all_s_s_mul_ext assms(1) case_prodI mem_Collect_eq mset_add multi_member_last) 
qed

lemma actop_size_ge_1_impl_s_mul_ackbo:
  assumes "(s,t) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "f \<in> AC"
    and "size (actop f t) > 1"
    and "\<And> s t. s \<rhd> t \<Longrightarrow> (s,t) \<in> s_rel"
    and "order_pair s_rel ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "({#s#}, actop f t) \<in> s_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) s_rel"
proof -
  have "size (actop f t) \<noteq> 1" using assms(3) by arith
  then have "\<And>b. b \<in># actop f t \<Longrightarrow> (\<exists>a. a \<in># {#t#} \<and> (a, b) \<in> s_rel)" using assms(4) actop_impl_subterm
    by (metis multi_member_last size_single)
  then have "({#t#}, actop f t) \<in> s_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) s_rel" 
    using s_mul_extI[of "{#t#}" "{#}" "{#t#}" "actop f t" "{#}" "actop f t" "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" "s_rel"]
    by (metis add_mset_not_empty all_s_s_mul_ext)  
  then show ?thesis using assms(5) by (metis (no_types, lifting) add_cancel_right_right assms(1) ns_mul_ext_singleton ns_s_mul_ext_sum_trans)  
qed

lemma aux2_for_ctxt_closure:
  assumes "ackbo s t"
    and "f \<in> AC"
  shows "({#s#}, actop f t) \<in> s_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) {(x,y). ackbo x y}"
proof (cases "size (actop f t) = 1")
  case True
  then have "actop f t = {#t#}" using actop_size_one by auto
  then show ?thesis by (simp add: assms(1)) 
next
  case False
  then have sm:"\<And>b. b \<in># actop f t \<Longrightarrow> (\<exists>a. a \<in># {#t#} \<and> ackbo a b)" using ackbo_subterm_property actop_impl_subterm
    by (metis multi_member_last size_single) 
  then have "({#t#}, actop f t) \<in> s_mul_ext ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) {(x,y). ackbo x y} " 
    using s_mul_extI[of "{#t#}" "{#}" "{#t#}" "actop f t" "{#}" "actop f t" "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" "{(x, y). ackbo x y}"] by auto 
  then show ?thesis using sm
    by (smt ackbo_ord_p.trans_S_point add_mset_not_empty all_s_s_mul_ext assms(1) case_prodI mem_Collect_eq mset_add multi_member_last single_eq_add_mset) 
qed

(*
    Section: Closure under context proofs
*)

lemma lemma_5_7:
  assumes "ackbo s t"
    and "h \<in> AC"
    and "length (bef @ aft) = 1"
    and "u \<in> set (bef @ aft)"
  shows "ackbo (Fun h (bef @ s # aft)) (Fun h (bef @ t # aft))"
proof -
  let ?S = "actop h s"
  let ?T = "actop h t"
  let ?U = "actop h u"
  let ?f = "(h,2)"
  let ?lhs = "Fun h (bef @ s # aft)"
  let ?rhs = "Fun h (bef @ t # aft)"
  let ?str = "{(x,y). ackbo x y}"
  let ?ns = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
  have v:"vars_term_ms t \<subseteq># vars_term_ms s" using ackbo.simps assms(1) by blast
  have one_empty:"bef = [] \<or> aft = []" using assms(3)
    by (metis One_nat_def Suc_length_conv append_eq_Cons_conv append_is_Nil_conv length_0_conv)
  then have one_con_u:"bef = [u] \<or> aft = [u]" using assms(3-4) by (smt List.self_append_conv One_nat_def Suc_length_conv append_self_conv2 in_set_simps(2) length_0_conv)  
  then have "Fun h (bef @ s # aft) = Fun h [u,s] \<or> Fun h (bef @ s # aft) = Fun h [s,u]" using one_empty by auto
  then have lhs:"actop h ?lhs = ?S + ?U" by auto
  have "Fun h (bef @ t # aft) = Fun h [u,t] \<or> Fun h (bef @ t # aft) = Fun h [t,u]" using one_empty one_con_u by auto
  then have rhs:"actop h ?rhs = ?T + ?U" by auto
  consider (a) "weight s > weight t" | (b) "weight s = weight t" using assms(1) ackbo_weight_pos by blast
  then show ?thesis
  proof cases
    case a
    then have "weight ((More h bef Hole aft)\<langle>s\<rangle>) > weight ((More h bef Hole aft)\<langle>t\<rangle>)" by auto
    moreover have "vars_term_ms ((More h bef Hole aft)\<langle>t\<rangle>) \<subseteq># vars_term_ms ((More h bef Hole aft)\<langle>s\<rangle>)" using assms(1) ackbo.simps by auto
    ultimately show ?thesis using case_w by fastforce
  next
    case b
    have vars:"vars_term_ms ?rhs \<subseteq># vars_term_ms ?lhs" using v by auto
    have wei:"weight ?lhs = weight ?rhs" using b by auto
    have sub:"?T \<restriction>\<^sub>n ?f + ?T\<restriction>\<^sub>v - ?S\<restriction>\<^sub>v \<subseteq># ?T" by (metis subset_rel_filterd_comb filter_minus_assoc)
    from b show ?thesis
    proof (cases "is_Var t")
      case True
      then have "ackbo_case0 s t" using b assms(1) by (simp add: ackbo_rhs_var_imp_case0)
      then obtain f s1 where term_s:"s = Fun f [s1]" using is_FunE by (meson ackbo_case0.simps)
      then have "w(f,1) = 0" using True b
        by (metis One_nat_def add_cancel_right_left eq_iff is_FunI le_add2 length_Cons list.size(3) min_weigth_of_not_empty_list not_Cons_self2 weight.elims weight.simps(2))
      then have pr:"pr_strict (f,1) (h,2)" using adm by auto
      then have top_s:"?S = {#s#}" using actop.simps(4) term_s by auto
      then have "?S \<restriction>\<^sub>n ?f = {#s#}" using pr pr_str term_s unfolding filter_fun_def by auto  
      moreover have top_t:"?T = {#t#}" using actop.simps(2) True by auto
      ultimately have "filtered_nless_rel_s ?S ?T ?f ?ns ?str " using assms(1) pr True filter_fun_of_var_empty top_s term_s sub by auto
      then have "ackbo_case3_filtered_rel (?S + ?U) (?T + ?U) ?f ?ns ?str"
        using ackbo_ord_p.order_pair_axioms ackbo_case3_closed_under_mset_union by blast
      then have "ackbo_case3 ackbo ?lhs ?rhs" using ackbo_case3.simps(1) assms lhs rhs
        by (smt Suc_eq_plus1 add_Suc_right case_prodI length_Cons length_append nat_1_add_1) 
      then show ?thesis using assms(1) ackbo.case_3 wei vars by blast
    next
      case False
      obtain g gs where term_t:"t = Fun g gs" using False by blast
      obtain f fs where term_s:"s = Fun f fs" using assms(1) by (meson ackbo_base_lhs_fun ackbo_ext_lhs_fun ackbo_eq_ackbo_wk is_FunE)
      then have inv:"pr_strict (f, length fs) (g, length gs) \<or> f = g \<and> length fs = length gs"
        using assms(1) term_t term_s ackbo_no_var_impl_prstrict_or_eq_root b by blast
      consider (i) "\<not> pr_strict (h,2) (f, length fs) \<and> (f \<noteq> h \<or> length fs \<noteq> 2)"|
        (ii)  "pr_strict (h,2) (g,length gs) \<and> f = h \<and> length fs = 2"|
        (iii) "f = h \<and> g = h \<and> length fs = length gs \<and> length fs = 2"|
        (iv)  "pr_strict (h,2) (g,length gs) \<and> pr_strict (h,2) (f,length fs)"
        using inv term_t term_s False by (smt pr_trans)
      then have "ackbo_case3_filtered_rel (?S + ?U) (?T + ?U) ?f ?ns ?str"
      proof cases
        case i
        then have top_s:"?S = {#s#}" using actop.simps term_s
          by (smt One_nat_def Suc_eq_plus1 actop_singleton length_Cons list.size(3) nat_1_add_1 term.sel(4))
        then have "?S \<restriction>\<^sub>n ?f = {#s#}" using i term_s unfolding filter_fun_def by auto
        then have "(?S \<restriction>\<^sub>n ?f, ?T) \<in> s_mul_ext ?ns ?str" using aux2_for_ctxt_closure assms(1-2) by auto
        then have "(?S \<restriction>\<^sub>n ?f, ?T \<restriction>\<^sub>n ?f + ?T\<restriction>\<^sub>v - ?S\<restriction>\<^sub>v) \<in> s_mul_ext ?ns ?str"
          using s_mul_ext_subset_trans ackbo_ord_p.order_pair_axioms sub by blast
        then have "filtered_nless_rel_s ?S ?T ?f ?ns ?str" by auto
        then show ?thesis  using ackbo_ord_p.order_pair_axioms ackbo_case3_closed_under_mset_union by blast
      next
        case ii
        then have "h \<noteq> g \<or> length gs \<noteq> 2" using ii pr_irr by auto
        then have top_t:"actop h t = {#Fun g gs#}" using term_t actop.simps
          by (smt One_nat_def Suc_eq_plus1 actop_singleton length_Cons list.size(3) nat_1_add_1 term.sel(4)) 
        then have emp:"?T\<restriction>\<^sub>n ?f = {#}" using term_t ii pr_str unfolding filter_fun_def by auto
        have "size ?S > 1" using ii term_s Bin_cases_with_length[of s] actop.simps(1)[of h f]
          by (smt actop_mim_size nat_less_le non_empty_plus_non_empty_not_single not_one_le_zero size_1_singleton_mset size_empty term.distinct(1) term.inject(2)) 
        then have "size ?S > size ?T" using top_t by auto
        then have "filtered_nless_rel_ns ?S ?T ?f ?ns ?str \<and> size ?S > size ?T" 
          using emp term_t top_t filter_var_of_fun_empty ns_mul_ext_bottom by auto
        then show ?thesis using ackbo_ord_p.order_pair_axioms ackbo_case3_closed_under_mset_union by blast 
      next
        case iii
        then have "ackbo_case3 ackbo s t" using assms(1-2) ackbo_acsym_imp_case3[of s t] b term_s term_t by auto
        then show ?thesis 
          using ackbo_ord_p.order_pair_axioms ackbo_case3_closed_under_mset_union term_s term_t ackbo_case3_Elim iii by blast 
      next
        case iv
        then have top_s:"?S = {#s#}" using pr_irr actop.simps term_s
          by (smt One_nat_def Suc_eq_plus1 actop_singleton length_Cons list.size(3) nat_1_add_1 term.sel(4))
        moreover have top_t:"?T = {#t#}" using pr_irr actop.simps term_t iv
          by (smt One_nat_def Suc_eq_plus1 actop_singleton length_Cons list.size(3) nat_1_add_1 term.sel(4))
        ultimately have "filtered_less_rel_s ?S ?T ?f ?ns ?str" using assms(1) iv term_s term_t
          by (smt add_cancel_left_right assms(2) aux2_for_ctxt_closure case_prodI filter_fun_add_mset_order filter_fun_dist mem_Collect_eq) 
        moreover have "size ?S = size ?T" using top_s top_t by auto
        moreover have "filtered_nless_rel_ns ?S ?T ?f ?ns ?str"
          using iv term_t top_t filter_var_of_fun_empty ns_mul_ext_bottom pr_str unfolding filter_fun_def by auto
        ultimately show ?thesis 
          using ackbo_ord_p.order_pair_axioms ackbo_case3_closed_under_mset_union by blast
      qed
      then have "ackbo_case3 ackbo ?lhs ?rhs" using ackbo_case3.simps(1) assms lhs rhs
        by (smt Suc_eq_plus1 add_Suc_right case_prodI length_Cons length_append nat_1_add_1) 
      then show ?thesis using assms(1) ackbo.case_3 wei vars by blast
    qed
  qed
qed


lemma ackbo_mono_one:
  assumes "ackbo s t"
  shows "ackbo (Fun f (ss1 @ s # ss2)) (Fun f (ss1 @ t # ss2))"
proof -
  let ?ss = "ss1 @ s # ss2"
  let ?ts = "ss1 @ t # ss2"
  let ?s = "Fun f ?ss"
  let ?t = "Fun f ?ts"
  have w: "weight t \<le> weight s" using ackbo_weight_eq assms by fastforce
  have v: "vars_term_ms t \<subseteq># vars_term_ms s" using assms ackbo.simps by auto
  have v': "vars_term_ms ?t \<subseteq># vars_term_ms ?s" using v by (induct ss1) auto
  have w': "weight ?t \<le> weight ?s" using w by (induct ss1) auto
  have eq_len:"length ?ss = length ?ts" by auto
  consider (a) "f \<in> AC \<and> length ?ss = 2" | (b) "f \<notin> AC \<or> length ?ss \<noteq> 2" by blast
  then show ?thesis
  proof cases
    case a
    then have "length (ss1 @ ss2) = 1" by auto
    then show ?thesis using lemma_5_7[of s t f ss1 ss2] a
      by (metis append_Nil assms in_set_conv_decomp length_0_conv list.exhaust zero_neq_one)  
  next
    case b
    have "\<forall> i < length ss1. (ss1 ! i, ss1 ! i) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" by blast
    then have "(\<exists> i < length ?ss. i < length ?ts \<and> (\<forall> j < i. (?ss ! j, ?ts ! j) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) \<and> ackbo (?ss ! i) (?ts !i))"
      using assms
      by (metis (no_types, lifting) List.self_append_conv append_Cons_nth_left le_add_same_cancel1 length_append length_greater_0_conv list.distinct(1) list_eq_iff_nth_eq nat_less_le nth_append_length)
    then have lex: "fst (lex_ext (\<lambda> x y. (ackbo x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length ?ss) ?ss ?ts)"
      using eq_len by (simp add: lex_ext_iff)
    then have "ackbo_case2 ackbo ?s ?t" using b ackbo_case2.simps(1) by auto
    then show ?thesis using w' v' by (metis ackbo.case_2 case_w le_neq_implies_less) 
  qed
qed

lemma ackbo_ctxt:
  assumes "ackbo s t"
  shows "ackbo (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
  using one_imp_ctxt_closed [of "{(x, y). ackbo x y}", to_pred, OF ackbo_mono_one, THEN ctxt.closedD, to_pred] assms by blast

(*
    Section: Closed under substitution proof
*)

abbreviation actop_mset
  where
    "actop_mset f M \<equiv> \<Union># image_mset (\<lambda> x. actop f x) M"

notation actop_mset ("\<nabla># _ _" [200, 200] 90)

lemma actop_impl_eq_or_s_mul:
  assumes "f \<in> AC"
    and "\<And> s t. s \<rhd> t \<Longrightarrow> (s,t) \<in> s_rel"
  shows "{#v#} = (actop f v) \<or> ({#v#}, actop f v) \<in> s_mul_ext ns_rel s_rel \<and> size {#v#} < size (actop f v)"
proof -
  consider (a) "\<exists> vs. v = Fun f vs \<and> length vs = 2"|
    (b) "\<exists> g vs. is_Var v \<or> v = Fun g vs \<and> (g \<noteq> f \<or> length vs \<noteq> 2)"
    by (metis is_FunE)
  then show ?thesis
  proof cases
    case a
    then obtain s1 s2 where term_v:"v = Fun f [s1, s2]" using Bin_cases_with_length by (metis term.distinct(1) term.sel(4))
    then have s:"size {#v#} < size (actop f v)" using actop.simps(1) actop_mim_size add_strict_increasing
      by (metis gr0I not_one_le_zero size_single size_union)
    have "\<And> x. x \<in># actop f v \<Longrightarrow> v \<rhd> x" by (simp add: actop_mset_elem_subterm term_v)
    then have "({#v#}, actop f v) \<in> s_mul_ext ns_rel s_rel" by (simp add: all_s_s_mul_ext assms(2))  
    then show ?thesis using s by auto
  next
    case b
    then show ?thesis
      by (metis (no_types, lifting) One_nat_def Suc_eq_plus1 Term.term.simps(2) actop_singleton is_FunI length_Cons list.size(3) nat_1_add_1) 
  qed
qed

lemma image_actop_impl_eq_or_s_mul:
  fixes V ::"('f, 'v) term multiset"
  assumes "f \<in> AC"
    and "\<And> s t. s \<rhd> t \<Longrightarrow> (s,t) \<in> s_rel"
    and "refl ns_rel"
  shows "V = \<nabla># f V \<or> (V, \<nabla># f V) \<in> s_mul_ext ns_rel s_rel \<and> size V < size (\<nabla># f V)"
proof (induct "size V" arbitrary: V)
  case 0
  then show ?case by auto
next
  case (Suc x)
  then have "\<exists> A v. V = A + {#v#}" by (metis add_mset_add_single size_eq_Suc_imp_eq_union)
  then obtain A v where V:" V = A + {#v#}" by blast
  let ?R = "A + {#v#}"
  let ?Eq_mset = "\<lambda> x. x = \<nabla># f x"
  let ?S_mul = "\<lambda> x. (x, \<nabla># f x) \<in> s_mul_ext ns_rel s_rel \<and> size x < size (\<nabla># f x)"
  let ?Conc = "\<lambda> x. ?Eq_mset x \<or> ?S_mul x" 
  have v:"?Conc {#v#}" using actop_impl_eq_or_s_mul assms by auto
  then have "x = size A" using Suc(2) V
    by (metis (no_types, lifting) Suc_eq_plus1 add_right_imp_eq size_single size_union)    
  then have "?Conc A" using Suc(1) by auto
  then consider (a) "?Eq_mset A"| (b) "?S_mul A"
    by blast
  then show ?case
  proof cases
    case a
    consider (i) "?Eq_mset {#v#}" | (ii) "?S_mul {#v#}" using v by blast
    then show ?thesis
    proof cases
      case i
      have "?Eq_mset (A + {#v#})"
        by (subst i, subst a) auto
      then show ?thesis using V by blast
    next
      case ii
      have size:"size (A + {#v#}) < size (\<nabla># f (A + {#v#}))" using a ii by auto
      then have "(A, \<nabla># f A) \<in> ns_mul_ext ns_rel s_rel" using a ns_mul_ext_refl assms(3) ii by auto
      then have "?S_mul (A + {#v#})" using ii a s_ns_mul_ext_union_compat size unfolding image_mset_union by fastforce
      then show ?thesis using V by auto
    qed
  next
    case b
    consider (i) "?Eq_mset {#v#}" | (ii) "?S_mul {#v#}" using v by blast
    then show ?thesis
    proof cases
      case i
      have size:"size (A + {#v#}) < size (\<nabla># f (A + {#v#}))" using b by (subst i) auto
      then have "({#v#}, \<nabla># f {#v#}) \<in> ns_mul_ext ns_rel s_rel" using ns_mul_ext_refl assms(3) by (subst i) auto
      then have "?S_mul (A + {#v#})" using i b s_ns_mul_ext_union_compat size unfolding image_mset_union by fastforce
      then show ?thesis using V by auto
    next
      case ii
      have size:"size (A + {#v#}) < size (\<nabla># f (A + {#v#}))" using b ii by auto
      then have "({#v#}, \<nabla># f {#v#}) \<in> ns_mul_ext ns_rel s_rel" using ii s_ns_mul_ext by auto
      then have "?S_mul (A + {#v#})" using b s_ns_mul_ext_union_compat size unfolding image_mset_union by (metis sum_mset.union) 
      then show ?thesis using V by auto
    qed
  qed
qed


lemma aux_filter_fun_actop_subst_eq_expr:
  fixes f :: 'f
  shows "filter_fun (actop f (s \<cdot> \<sigma>)) R (f, n) = filter_fun (actop f s) R (f, n) \<cdot># \<sigma> 
                  + filter_fun (\<nabla># f (actop f s\<restriction>\<^sub>v \<cdot># \<sigma>)) R (f, n)"
proof (induct s)
  case (Var x)
  then have "actop f (Var x) = {#Var x#}" using actop.simps(2) by auto
  have "image_mset (\<lambda> x. actop f (x \<cdot> \<sigma>)) (actop f (Var x)) = {#actop f ((Var x) \<cdot> \<sigma>)#}" by auto
  moreover have "\<Union># {#actop f ((Var x) \<cdot> \<sigma>)#} = actop f ((Var x) \<cdot> \<sigma>)" by auto
  ultimately have "(\<Union># image_mset (\<lambda> x. actop f (x \<cdot> \<sigma>)) (actop f (Var x)))\<restriction>\<^sub>n (f, length (args s)) =
                   (actop f ((Var x) \<cdot> \<sigma>))\<restriction>\<^sub>n (f, length (args s))"
    by auto
  then show ?case by (simp add: filter_fun_def)
next
  case (Fun g gs)
  consider (a) "g = f \<and> length gs = 2" | (b) "g \<noteq> f \<or> length gs \<noteq> 2" by blast
  then show ?case
  proof cases
    case a
    then obtain g1 g2 where sub_gs:"gs = [g1,g2]" using length_0_conv Suc_length_conv by (smt numeral_2_eq_2)
    then have actop_sub:"actop f (Fun g gs \<cdot> \<sigma>) = actop f (g1 \<cdot> \<sigma>) + actop f (g2 \<cdot> \<sigma>)" using actop.simps(1) sub_gs a by auto
    then have "filter_fun (actop f (g1 \<cdot> \<sigma>)) R (f, n) + filter_fun (actop f (g2 \<cdot> \<sigma>)) R (f, n) =
      filter_fun (actop f g1) R (f, n) \<cdot># \<sigma> + filter_fun (actop f g2) R (f, n) \<cdot># \<sigma> 
      + filter_fun (\<nabla># f (filter_var (actop f g1) \<cdot># \<sigma>)) R (f, n)
      + filter_fun (\<nabla># f (filter_var (actop f g2) \<cdot># \<sigma>)) R (f, n)" using Fun sub_gs by auto
    moreover have "actop f (Fun g gs) = actop f (g1) + actop f (g2)" using actop.simps(1) sub_gs a by auto
    ultimately show ?thesis using sub_gs actop_sub by (simp add: filter_fun_def)
  next
    case b
    have sub:"Fun g gs \<cdot> \<sigma> = Fun g (map (\<lambda> t. t \<cdot> \<sigma>) gs)" by auto
    then have "actop f (Fun g gs \<cdot> \<sigma>) = {#Fun g gs \<cdot> \<sigma>#}" using b
      by (metis (no_types, lifting) One_nat_def Suc_eq_plus1 actop_singleton length_Cons length_map list.size(3) nat_1_add_1 term.inject(2))
    moreover have "filter_fun (\<nabla># f (actop f (Fun g gs)\<restriction>\<^sub>v \<cdot># \<sigma>)) R (f, n) = {#}" using b filter_fun_def
      by (smt One_nat_def Suc_eq_plus1 actop_singleton filter_single_mset image_mset_is_empty_iff is_FunI length_Cons list.size(3) multiset_filter_subset nat_1_add_1 subset_mset.le_zero_eq sum_mset.empty term.inject(2)) 
    ultimately show ?thesis using filter_fun_def
      by (smt One_nat_def Suc_eq_plus1 Term.term.simps(6) actop_singleton add.right_neutral b filter_single_mset image_mset_is_empty_iff image_mset_single length_Cons length_map list.size(3) nat_1_add_1 sub term.inject(2)) 
  qed
qed

lemma aux_filter_var_actop_subst_eq_expr:
  shows "actop f (s \<cdot> \<sigma>)\<restriction>\<^sub>v = (\<nabla># f (actop f s\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v"
proof (induct s)
  case (Var x)
  then show ?case by auto
next
  case (Fun g gs)
  consider (a) "g = f \<and> length gs = 2" | (b) "g \<noteq> f \<or> length gs \<noteq> 2" by blast
  then show ?case
  proof cases
    case a
    then obtain g1 g2 where sub_gs:"gs = [g1,g2]"
      by (smt One_nat_def Suc_eq_plus1 Suc_length_conv length_0_conv nat_1_add_1) 
    then have actop_sub:"actop f (Fun g gs \<cdot> \<sigma>) = actop f (g1 \<cdot> \<sigma>) + actop f (g2 \<cdot> \<sigma>)" using actop.simps(1) sub_gs a by auto
    moreover have "filter_var (actop f (g1 \<cdot> \<sigma>) + actop f (g2 \<cdot> \<sigma>)) = filter_var (actop_mset f (filter_var (actop f g2) \<cdot># \<sigma>))
                   + filter_var (actop_mset f (filter_var (actop f g1) \<cdot># \<sigma>))" using Fun sub_gs by auto
    moreover have "actop f (Fun g gs) = actop f (g1) + actop f (g2)" using actop.simps(1) sub_gs a by auto
    ultimately show ?thesis by auto
  next
    case b
    have sub:"Fun g gs \<cdot> \<sigma> = Fun g (map (\<lambda> t. t \<cdot> \<sigma>) gs)" by auto
    then have "actop f (Fun g gs \<cdot> \<sigma>) = {#Fun g gs \<cdot> \<sigma>#}" using b
      by (metis (no_types, lifting) One_nat_def Suc_eq_plus1 actop_singleton length_Cons length_map list.size(3) nat_1_add_1 term.inject(2))
    moreover have "actop f (Fun g gs) = {#Fun g gs#}" using b
      by (metis (mono_tags, lifting) One_nat_def Suc_eq_plus1 actop_singleton length_Cons list.size(3) nat_1_add_1 term.inject(2)) 
    ultimately show ?thesis by simp
  qed
qed

lemma tests_subs:
  shows "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v - (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v \<subseteq># (\<nabla># f ((T - S)\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v"
proof -
  have "(\<nabla># f ((T - S)\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v = (\<nabla># f ((T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<cdot># \<sigma>))\<restriction>\<^sub>v" by simp
  moreover have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma> - S\<restriction>\<^sub>v \<cdot># \<sigma> ))\<restriction>\<^sub>v \<subseteq># (\<nabla># f ((T\<restriction>\<^sub>v - S\<restriction>\<^sub>v) \<cdot># \<sigma>))\<restriction>\<^sub>v"
    by (metis (no_types) image_mset_union multiset_filter_mono subset_eq_diff_conv subset_mset.dual_order.refl subset_mset.le_iff_add sum_mset.union)
  ultimately show ?thesis
  proof -
    have "\<Union>#image_mset (actop f) (T \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>v \<subseteq># \<Union>#image_mset (actop f) ((T - S + S) \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>v"
      by (metis (no_types) filter_union_mset image_mset_union subset_eq_diff_conv subset_mset.dual_order.refl subset_mset.le_iff_add sum_mset.union)
    then show ?thesis
      by (simp add: subset_eq_diff_conv)
  qed
qed

lemma more_subs:
  shows "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,n) \<subseteq># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,n) + (\<nabla># f ((T - S)\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,n)"
proof -
  have "T\<restriction>\<^sub>v \<cdot># \<sigma> \<subseteq># S\<restriction>\<^sub>v \<cdot># \<sigma> + (T - S)\<restriction>\<^sub>v \<cdot># \<sigma>"
    by (metis (no_types, lifting) Diff_eq_empty_iff_mset diff_add filter_diff_mset diff_cancel image_mset_subseteq_mono image_mset_union)
  then have "\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>) \<subseteq># \<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma> + (T - S)\<restriction>\<^sub>v \<cdot># \<sigma>)"
    by (metis image_mset_union subset_mset.le_iff_add sum_mset.union)
  then have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,n) \<subseteq># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma> + (T - S)\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,n)"
    by (metis filter_fun_def multiset_filter_mono)
  then show ?thesis by (simp add: filter_fun_dist) 
qed

lemma filter_nless_presv_elm:
  assumes "S\<restriction>\<^sub>p (f,2) \<noteq> {#}"
  shows "(\<nabla># f S)\<restriction>\<^sub>p (f,2) \<noteq> {#}"
proof -
  from assms have "\<exists> s U. add_mset s U = S \<and> s \<in># S \<restriction>\<^sub>p (f,2)" unfolding filter_fun_def
    by (meson insert_DiffM mset_subset_eqD multiset_filter_subset multiset_nonemptyE)  
  then obtain s U g hs where w:"add_mset s U = S \<and> s = Fun g hs \<and> s \<in># S \<restriction>\<^sub>p (f,2)" unfolding filter_fun_def
    by (metis (no_types, lifting) filter_mset_eq_conv is_Fun_Fun_conv term.case_eq_if)
  then have "pr_strict (f,2) (g, length hs)" unfolding filter_fun_def by auto
  then have "length hs \<noteq> 2 \<or> g \<noteq> f" using pr_irr by auto 
  then have "\<nabla># f S = {#s#} + \<nabla># f U" using w
    by (smt One_nat_def actop.simps(1) actop_non_Bin add.right_neutral add_Suc_right length_Cons list.size(3) nat_1_add_1 sum_mset.insert)
  moreover have "({#s#} + \<nabla># f U)\<restriction>\<^sub>p (f,2) \<noteq> {#}" by (metis (no_types, lifting) add_cancel_right_left empty_not_add_mset filter_fun_def filter_mset_add_mset filter_mset_eq_conv union_mset_add_mset_left w) 
  ultimately show ?thesis by auto  
qed

lemma aux_subst_closed_subset_rel:
  assumes "((T - S)\<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>p (f,2) \<noteq> {#}"
  shows "T\<restriction>\<^sub>n (f,2) \<cdot># \<sigma> + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,2) + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v - (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v \<subset># T\<restriction>\<^sub>n (f,2) \<cdot># \<sigma> + (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,2) + \<nabla># f ((T - S)\<restriction>\<^sub>v \<cdot># \<sigma>)"
proof -
  let ?f = "(f,2)"
  let ?U = "(T - S)\<restriction>\<^sub>v"
  have not_e:"(\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>p ?f \<noteq> {#}" using filter_nless_presv_elm assms by auto
  have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f \<subseteq># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>n ?f" using more_subs by blast
  then have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f \<subset># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>p ?f" using not_e
    by (metis add.commute subset_mset.add_strict_increasing subset_mset.gr_zeroI) 
  then have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v \<subset># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>p ?f"
    by (simp add: union_commute)
  then have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v \<subset># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)" 
    using filter_unions_eq[of "\<nabla># f (?U \<cdot># \<sigma>)" ?f] by (metis union_assoc)
  then have "T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v \<subset># T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)" by simp
  then have "T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v - (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v \<subset># T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)"
    using tests_subs[of f \<sigma> T S] by (smt cancel_ab_semigroup_add_class.diff_right_commute diff_union_cancelL subset_eq_diff_conv subset_mset.dual_order.strict_trans subset_mset_def union_commute) 
  then show ?thesis by blast
qed

lemma aux_subst_closed_subseteq_rel:
  shows "T\<restriction>\<^sub>n (f,n) \<cdot># \<sigma> + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,n) + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v - (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v \<subseteq># T\<restriction>\<^sub>n (f,n) \<cdot># \<sigma> + (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n (f,n) + \<nabla># f ((T - S)\<restriction>\<^sub>v \<cdot># \<sigma>)"
proof -
  let ?f = "(f,n)"
  let ?U = "(T - S)\<restriction>\<^sub>v"
  have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f \<subseteq># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>n ?f" using more_subs by blast
  then have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f \<subseteq># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>p ?f"
    by (metis (no_types, lifting) Diff_eq_empty_iff_mset Multiset.diff_add mset_subset_eq_add_left) 
  then have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v \<subseteq># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>p ?f"
    by (simp add: union_commute)
  then have "(\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v \<subseteq># (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)" 
    using filter_unions_eq[of "\<nabla># f (?U \<cdot># \<sigma>)" ?f] by (metis union_assoc)
  then have "T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?U \<cdot># \<sigma>))\<restriction>\<^sub>v \<subseteq># T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)" by simp
  then have "T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v - (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v \<subseteq># T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)"
    using tests_subs[of f \<sigma> T S] apply (auto)
    by (smt Multiset.diff_right_commute \<open>T \<restriction>\<^sub>n (f, n) \<cdot># \<sigma> + \<Union>#image_mset (actop f) (T \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>n (f, n) + \<Union>#image_mset (actop f) ((T - S) \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>v \<subseteq># T \<restriction>\<^sub>n (f, n) \<cdot># \<sigma> + \<Union>#image_mset (actop f) (S \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>n (f, n) + \<Union>#image_mset (actop f) ((T - S) \<restriction>\<^sub>v \<cdot># \<sigma>)\<close> add.commute diff_union_cancelL filter_diff_mset subset_eq_diff_conv subset_mset.add_diff_inverse subset_mset.add_increasing2 subset_mset.le_add_same_cancel2)
  then show ?thesis by blast
qed

lemma mset_subst_size[simp]:
  shows "size (U \<cdot># \<sigma>) = size U"
  using size_image_mset by auto

lemma filter_var_less_nless_pres_mset_size:
  shows "size (U\<restriction>\<^sub>n f) + size (U\<restriction>\<^sub>p f) + size (U\<restriction>\<^sub>v) = size U"
  by (metis (no_types, lifting) add.commute filter_unions_eq size_union union_assoc)

lemma subst_actop_size_eq:
  shows "size (actop f (s \<cdot> \<sigma>)) = size (actop f s\<restriction>\<^sub>n h) + size (actop f s\<restriction>\<^sub>p h) + size (\<nabla># f (actop f s\<restriction>\<^sub>v \<cdot># \<sigma>))"
proof (induct s)
  case (Var x)
  then show ?case by (simp add: filter_fun_def) 
next
  case (Fun g gs)
  then show ?case
  proof (cases "g = f \<and> length gs = 2")
    case True
    then have "\<exists> g1 g2. gs = [g1, g2]" using Bin_cases_with_length by (smt One_nat_def Suc_eq_plus1 Suc_length_conv length_0_conv nat_1_add_1)
    then obtain g1 g2 where w:"gs = [g1, g2]" by blast
    then have "actop f (Fun g gs \<cdot> \<sigma>) = actop f (g1 \<cdot> \<sigma>) + actop f (g2 \<cdot> \<sigma>)" using True actop.simps(1)[of f g "g1 \<cdot> \<sigma>" "g2 \<cdot> \<sigma>"] subst_apply_term.simps(2) by auto
    then have "size (actop f (Fun g gs \<cdot> \<sigma>)) = size (actop f (g1 \<cdot> \<sigma>)) + size (actop f (g2 \<cdot> \<sigma>))" by auto
    moreover have "size (actop f (Fun g gs)\<restriction>\<^sub>n h) +  size (actop f (Fun g gs)\<restriction>\<^sub>p h) + size (\<nabla># f (actop f (Fun g gs)\<restriction>\<^sub>v \<cdot># \<sigma>)) =
          size (actop f g1\<restriction>\<^sub>n h) + size (actop f g1\<restriction>\<^sub>p h) + size (\<nabla># f (actop f g1\<restriction>\<^sub>v \<cdot># \<sigma>)) +
          size (actop f g2\<restriction>\<^sub>n h) + size (actop f g2\<restriction>\<^sub>p h) + size (\<nabla># f (actop f g2\<restriction>\<^sub>v \<cdot># \<sigma>))"
      using w True actop.simps(1)[of f g g1 g2] subst_apply_term.simps(2) unfolding filter_fun_def by simp
    ultimately show ?thesis using Fun(1) w by simp 
  next
    case False
    then have act:"actop f (Fun g gs) = {#Fun g gs#}" by (metis (no_types, lifting) One_nat_def Suc_eq_plus1 actop_singleton length_Cons list.size(3) one_add_one term.inject(2)) 
    from False have "size (actop f (Fun g gs \<cdot> \<sigma>)) = 1" by (smt One_nat_def Suc_eq_plus1 actop_singleton length_Cons length_map list.size(3) nat_1_add_1 size_single subst_apply_term.simps(2) term.inject(2)) 
    moreover have "size (\<nabla># f (actop f (Fun g gs)\<restriction>\<^sub>v \<cdot># \<sigma>)) = 0" by (smt False One_nat_def Suc_eq_plus1 actop_singleton filter_single_mset image_mset_is_empty_iff is_FunI length_Cons list.size(3) nat_1_add_1 size_eq_0_iff_empty sum_mset.empty term.inject(2)) 
    moreover have "size (actop f (Fun g gs) \<restriction>\<^sub>n h) + size (actop f (Fun g gs) \<restriction>\<^sub>p h) = 1" using act by (metis add_cancel_right_right filter_single_mset filter_var_less_nless_pres_mset_size is_FunI size_empty size_single) 
    ultimately show ?thesis by simp
  qed
qed

lemma fi:
  shows "\<And> s. s \<in> set_mset (filter_fun S R f) \<Longrightarrow> s \<in> set_mset S"
  by (simp add: filter_fun_def)

lemma lemma_5_8_1:
  fixes s ::"('f, 'v) term"
    and \<sigma> ::"'v \<Rightarrow> ('f, 'v) term"
  assumes "filtered_nless_rel_s (actop f s) (actop f t) (f, 2) ns_rel s_rel"
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> ns_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> ns_rel"
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> s_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> s_rel"
    and "order_pair s_rel ns_rel"
    and "f \<in> AC"
    and "\<And> s t. s \<rhd> t \<Longrightarrow> (s,t) \<in> s_rel"
  shows "filtered_nless_rel_s (actop f (s \<cdot> \<sigma>)) (actop f (t \<cdot> \<sigma>)) (f, 2) ns_rel s_rel"
proof -
  let ?S = "actop f s"
  let ?T = "actop f t"
  let ?S' = "actop f (s \<cdot> \<sigma>)"
  let ?T' = "actop f (t \<cdot> \<sigma>)"
  let ?f = "(f,2)"
  let ?U = "(?T - ?S)\<restriction>\<^sub>v"
  have ref:"refl ns_rel" using assms(4) unfolding order_pair_def pre_order_pair_def by simp
  have eq:"?T\<restriction>\<^sub>v - ?S\<restriction>\<^sub>v = ?U" by auto 
  have wit:"(?S\<restriction>\<^sub>n ?f \<cdot># \<sigma>,(?T\<restriction>\<^sub>n ?f + ?U) \<cdot># \<sigma>) \<in> s_mul_ext ns_rel s_rel"
    using s_mul_ext_pres_sub_closer[OF assms(1-3)] by auto
      (smt fi filter_minus_assoc image_mset_union mset_subset_eqD subset_rel_filterd_comb)  
  have s'_eq:"?S'\<restriction>\<^sub>n ?f = ?S\<restriction>\<^sub>n ?f  \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f" using aux_filter_fun_actop_subst_eq_expr by blast
  have "(?S\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + ?U \<cdot># \<sigma>) \<in> s_mul_ext ns_rel s_rel"
    using wit s_mul_ext_union_compat[of _ _ ns_rel s_rel "(\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f"] assms(3) ref by (smt image_mset_union r_lr union_assoc union_commute) 
  moreover have "?U \<cdot># \<sigma> = \<nabla># f (?U \<cdot># \<sigma>) \<or> (?U \<cdot># \<sigma>, \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ns_rel s_rel"
    using image_actop_impl_eq_or_s_mul[OF assms(5,6) ref] by blast
  ultimately have w:"(?S\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ns_rel s_rel"
    by (smt assms(4) s_ns_mul_ext s_ns_mul_ext_sum_trans)
  have "?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + (\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v - (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>v \<subseteq># ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)"
    using aux_subst_closed_subseteq_rel by blast
  then have "(?S'\<restriction>\<^sub>n ?f, ?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v) \<in> s_mul_ext ns_rel s_rel" 
    using s'_eq w aux_filter_fun_actop_subst_eq_expr aux_filter_var_actop_subst_eq_expr
  proof -
    have f1: "actop f (t \<cdot> \<sigma>) \<restriction>\<^sub>n (f, 2) + actop f (t \<cdot> \<sigma>) \<restriction>\<^sub>v - actop f (s \<cdot> \<sigma>) \<restriction>\<^sub>v \<subseteq># actop f t \<restriction>\<^sub>n (f, 2) \<cdot># \<sigma> + \<Union>#image_mset (actop f) (actop f s \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>n (f, 2) + \<Union>#image_mset (actop f) ((actop f t - actop f s) \<restriction>\<^sub>v \<cdot># \<sigma>)"
      by (metis \<open>actop f t \<restriction>\<^sub>n (f, 2) \<cdot># \<sigma> + \<Union>#image_mset (actop f) (actop f t \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>n (f, 2) + \<Union>#image_mset (actop f) (actop f t \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>v - \<Union>#image_mset (actop f) (actop f s \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>v \<subseteq># actop f t \<restriction>\<^sub>n (f, 2) \<cdot># \<sigma> + \<Union>#image_mset (actop f) (actop f s \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>n (f, 2) + \<Union>#image_mset (actop f) ((actop f t - actop f s) \<restriction>\<^sub>v \<cdot># \<sigma>)\<close> aux_filter_fun_actop_subst_eq_expr aux_filter_var_actop_subst_eq_expr)
    have "(actop f (s \<cdot> \<sigma>) \<restriction>\<^sub>n (f, 2), actop f t \<restriction>\<^sub>n (f, 2) \<cdot># \<sigma> + \<Union>#image_mset (actop f) (actop f s \<restriction>\<^sub>v \<cdot># \<sigma>) \<restriction>\<^sub>n (f, 2) + \<Union>#image_mset (actop f) ((actop f t - actop f s) \<restriction>\<^sub>v \<cdot># \<sigma>)) \<in> s_mul_ext ns_rel s_rel"
      by (metis aux_filter_fun_actop_subst_eq_expr w)
    then show ?thesis
      using f1 assms(4) s_mul_ext_subset_trans by blast
  qed
  then show ?thesis by blast
qed

lemma lemma_5_8_2:
  fixes s ::"('f, 'v) term"
    and \<sigma> ::"'v \<Rightarrow> ('f, 'v) term"
  assumes "filtered_nless_rel_ns (actop f s) (actop f t) (f, 2) ns_rel s_rel" (is "filtered_nless_rel_ns ?S ?T ?f ?ac_rel s_rel")
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> ns_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> ns_rel" 
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> s_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> s_rel"
    and "order_pair s_rel ns_rel"
    and "f \<in> AC"
    and "\<And> s t. s \<rhd> t \<Longrightarrow> (s,t) \<in> s_rel"
    and "\<not> filtered_nless_rel_s (actop f (s \<cdot> \<sigma>)) (actop f (t \<cdot> \<sigma>)) (f, 2) ns_rel s_rel"
  shows "filtered_nless_rel_ns (actop f (s \<cdot> \<sigma>)) (actop f (t \<cdot> \<sigma>)) (f, 2) ns_rel s_rel
          \<and> ((actop f t - actop f s)\<restriction>\<^sub>v \<cdot># \<sigma>)\<restriction>\<^sub>p (f,2) = {#} \<and> (actop f t - actop f s)\<restriction>\<^sub>v \<cdot># \<sigma> = \<nabla># f ((actop f t - actop f s)\<restriction>\<^sub>v \<cdot># \<sigma>)"
proof -
  let ?S' = "actop f (s \<cdot> \<sigma>)"
  let ?T' = "actop f (t \<cdot> \<sigma>)"
  let ?U = "(?T - ?S)\<restriction>\<^sub>v"
  have ref:"refl ?ac_rel" using assms(4) unfolding order_pair_def pre_order_pair_def by simp
  have "(?S\<restriction>\<^sub>n ?f \<cdot># \<sigma>,(?T\<restriction>\<^sub>n ?f + ?U) \<cdot># \<sigma>) \<in> ns_mul_ext ?ac_rel s_rel"
    using ns_mul_ext_pres_sub_closer[OF assms(1-3)]
    by (smt fi filter_diff_mset filter_minus_assoc mset_subset_eqD subset_rel_filterd_comb) 
  then have up:"(?S\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + ?U \<cdot># \<sigma>) \<in> ns_mul_ext ?ac_rel s_rel"
    using  ns_mul_ext_union_compat[of _ _ ?ac_rel s_rel "(\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f"] assms(3) ref by (smt image_mset_union r_lr union_assoc union_commute) 
  consider (a) "?U \<cdot># \<sigma> = \<nabla># f (?U \<cdot># \<sigma>)"|
    (b) "?U \<cdot># \<sigma> \<noteq> \<nabla># f (?U \<cdot># \<sigma>)" by blast
  then have w:"(?S'\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)) \<in> ns_mul_ext ?ac_rel s_rel"
  proof cases
    case a
    then show ?thesis using up by (simp add: aux_filter_fun_actop_subst_eq_expr)
  next
    case b
    then have "(?U \<cdot># \<sigma>, \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ?ac_rel s_rel" using image_actop_impl_eq_or_s_mul[OF assms(5,6) ref] by blast 
    then have "(?S\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ?ac_rel s_rel"
      using up ns_s_mul_ext_sum_trans s_ns_mul_ext by (metis (no_types, lifting) assms(4) union_commute) 
    then have s_rel:"(?S'\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ?ac_rel s_rel"
      by (simp add: aux_filter_fun_actop_subst_eq_expr) 
    then show ?thesis using s_ns_mul_ext[OF s_rel] by simp
  qed
  consider (a) "?U \<cdot># \<sigma> = \<nabla># f (?U \<cdot># \<sigma>) \<and> (?U \<cdot># \<sigma>)\<restriction>\<^sub>p (f,2) = {#}"|
    (b) "?U \<cdot># \<sigma> \<noteq> \<nabla># f (?U \<cdot># \<sigma>)" | (c) "(?U \<cdot># \<sigma>)\<restriction>\<^sub>p (f,2) \<noteq> {#}" by blast
  then show ?thesis
  proof cases
    case a
    have sub:"?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v \<subseteq># ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)"
      using aux_subst_closed_subseteq_rel aux_filter_fun_actop_subst_eq_expr by (metis aux_filter_var_actop_subst_eq_expr)
    then have "(?S'\<restriction>\<^sub>n ?f, ?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v) \<in> ns_mul_ext ?ac_rel s_rel" 
      using ns_mul_ext_subset_trans[OF assms(4) w sub] by auto
    then show ?thesis using a by simp
  next
    case b
    then have "(?U \<cdot># \<sigma>, \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ?ac_rel s_rel" using image_actop_impl_eq_or_s_mul[OF assms(5,6) ref] by blast 
    then have "(?S\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ?ac_rel s_rel"
      using up ns_s_mul_ext_sum_trans s_ns_mul_ext by (metis (no_types, lifting) assms(4) union_commute)
    then have w:"(?S'\<restriction>\<^sub>n ?f, ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)) \<in> s_mul_ext ?ac_rel s_rel" by (simp add: aux_filter_fun_actop_subst_eq_expr) 
    have sub:"?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v \<subseteq># ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)"
      using aux_subst_closed_subseteq_rel aux_filter_fun_actop_subst_eq_expr by (metis aux_filter_var_actop_subst_eq_expr)
    then have "(?S'\<restriction>\<^sub>n ?f, ?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v) \<in> s_mul_ext ?ac_rel s_rel" 
      using s_mul_ext_subset_trans[OF assms(4) w sub] by auto
    then show ?thesis using assms(7) by blast 
  next 
    case c
    then have sub:"?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v \<subset># ?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>)"
      using aux_subst_closed_subset_rel aux_filter_fun_actop_subst_eq_expr by (metis aux_filter_var_actop_subst_eq_expr)
    then have s_mul:"(?T\<restriction>\<^sub>n ?f \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>n ?f + \<nabla># f (?U \<cdot># \<sigma>), ?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v) \<in> s_mul_ext ?ac_rel s_rel"
      using supset_imp_s_mul_ext[OF ref sub] by simp
    then have "(?S'\<restriction>\<^sub>n ?f, ?T'\<restriction>\<^sub>n ?f + ?T'\<restriction>\<^sub>v - ?S'\<restriction>\<^sub>v) \<in> s_mul_ext ?ac_rel s_rel" 
      using ns_s_mul_ext_sum_trans[OF assms(4) _ s_mul] w by (metis add.right_neutral) 
    then show ?thesis using assms(7) by simp
  qed
qed

abbreviation int_size
  where "int_size M \<equiv> int (size M)"

lemma lemma_5_8_3:
  fixes s ::"('f, 'v) term"
    and \<sigma> ::"'v \<Rightarrow> ('f, 'v) term"
  assumes "filtered_nless_rel_ns (actop f s) (actop f t) (f, 2) ns_rel s_rel" (is "filtered_nless_rel_ns ?S ?T ?f ?ac_rel s_rel")
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> ns_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> ns_rel" 
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> s_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> s_rel"
    and "order_pair s_rel ns_rel"
    and "f \<in> AC"
    and "\<And> s t. s \<rhd> t \<Longrightarrow> (s,t) \<in> s_rel"
    and "\<not> filtered_nless_rel_s (actop f (s \<cdot> \<sigma>)) (actop f (t \<cdot> \<sigma>)) (f, 2) ns_rel s_rel"
  shows "int_size (actop f s) - int_size (actop f t) \<le> int_size (actop f (s \<cdot> \<sigma>)) - int_size (actop f (t \<cdot> \<sigma>))"
proof -
  let ?S' = "actop f (s \<cdot> \<sigma>)"
  let ?T' = "actop f (t \<cdot> \<sigma>)"
  let ?U = "(?T - ?S)\<restriction>\<^sub>v"
  let ?X = "?S\<restriction>\<^sub>v \<inter># ?T\<restriction>\<^sub>v"
  have ref: "refl ns_rel" using assms(4) unfolding order_pair_def pre_order_pair_def by blast
  have u_subst_eq_actop:"?U \<cdot># \<sigma> = \<nabla># f (?U \<cdot># \<sigma>)" using lemma_5_8_2[OF assms(1-7)] by blast 
  have u:"?U = ?T\<restriction>\<^sub>v - ?X" by auto
  have "\<exists> H. ?S\<restriction>\<^sub>v = H + ?X" by (metis subset_mset.diff_add subset_mset.inf_le1)
  then obtain H where wit:"?S\<restriction>\<^sub>v = H + ?X" by blast
  have "int_size ?S' = int_size (?S\<restriction>\<^sub>n ?f) + int_size (?S\<restriction>\<^sub>p ?f) + int_size (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))"
    using subst_actop_size_eq by (metis (no_types) of_nat_add subst_actop_size_eq)
  moreover have "int_size (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>)) = int_size (\<nabla># f (H \<cdot># \<sigma>)) + int_size (\<nabla># f (?X \<cdot># \<sigma>))"
    using wit sub_mset_dist[of \<sigma> H ?X] image_mset_union by auto
  ultimately have size_S':"int_size ?S' = int_size (?S\<restriction>\<^sub>n ?f) + int_size (?S\<restriction>\<^sub>p ?f) + int_size (\<nabla># f (H \<cdot># \<sigma>)) + int_size (\<nabla># f (?X \<cdot># \<sigma>))"
    by auto
  have "int_size (H \<cdot># \<sigma>) \<le> int_size (\<nabla># f (H \<cdot># \<sigma>))" using image_actop_impl_eq_or_s_mul[OF assms(5,6) ref]
    by (metis dual_order.refl nat_less_le size_image_mset of_nat_le_iff)
  moreover have "int_size ?S = int_size (?S\<restriction>\<^sub>n ?f) + int_size (?S\<restriction>\<^sub>p ?f) + int_size H + int_size ?X"
    using wit by (smt filter_var_less_nless_pres_mset_size of_nat_add size_union)
  moreover have "size (H \<cdot># \<sigma>) \<le> size (\<nabla># f (H \<cdot># \<sigma>))" using image_actop_impl_eq_or_s_mul[OF assms(5,6) ref]
    by (metis dual_order.refl nat_less_le size_image_mset)
  ultimately have lhs:"int_size ?S' \<ge> int_size ?S - int_size ?X + int_size (\<nabla># f (?X \<cdot># \<sigma>))"
    using size_S' by auto
  have "int_size (\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>)) = int_size (\<nabla># f (?U \<cdot># \<sigma>)) + int_size (\<nabla># f (?X \<cdot># \<sigma>))"
    using u by (metis (no_types, lifting) image_mset_union image_mset_union size_union subset_mset.diff_add subset_mset.le_inf_iff subset_mset.order_refl sum_mset.union of_nat_add) 
  moreover have "int_size (\<nabla># f (?U \<cdot># \<sigma>)) = int_size (?U \<cdot># \<sigma>)" using u_subst_eq_actop by auto
  moreover have "int_size (?U \<cdot># \<sigma>) = int_size (?T\<restriction>\<^sub>v) - int_size ?X"
    by (simp add: multiset_inter_commute size_Diff_subset_Int of_nat_diff size_mset_mono)
  moreover have "int_size ?T' = int_size (?T\<restriction>\<^sub>n ?f) + int_size (?T\<restriction>\<^sub>p ?f) + int_size (\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>))"
    using subst_actop_size_eq by (metis (no_types) of_nat_add subst_actop_size_eq)
  ultimately have "int_size ?T - int_size ?X + int_size (\<nabla># f (?X \<cdot># \<sigma>)) = int_size ?T'"
    by (smt filter_var_less_nless_pres_mset_size of_nat_add) 
  then show ?thesis using lhs by linarith 
qed

lemma lemma_5_8_4:
  fixes s ::"('f, 'v) term"
    and \<sigma> ::"'v \<Rightarrow> ('f, 'v) term"
  assumes "filtered_less_rel_s (actop f s) (actop f t) (f, 2) ns_rel s_rel" (is "filtered_less_rel_s ?S ?T ?f ?ac_rel s_rel")
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> ns_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> ns_rel" 
    and "\<And>u v. u \<in> set_mset (actop f s) \<Longrightarrow> v \<in> set_mset (actop f t) \<Longrightarrow> (u,v) \<in> s_rel \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> s_rel"
    and "order_pair s_rel ns_rel"
    and "((actop f t - actop f s)\<restriction>\<^sub>v \<cdot># \<sigma>)\<restriction>\<^sub>p (f,2) = {#}"
    and "(actop f t - actop f s)\<restriction>\<^sub>v \<cdot># \<sigma> = \<nabla># f ((actop f t - actop f s)\<restriction>\<^sub>v \<cdot># \<sigma>)"
  shows "filtered_less_rel_s (actop f (s \<cdot> \<sigma>)) (actop f (t \<cdot> \<sigma>)) (f, 2) ns_rel s_rel"
proof -
  let ?S' = "actop f (s \<cdot> \<sigma>)"
  let ?T' = "actop f (t \<cdot> \<sigma>)"
  let ?U = "(?T - ?S)\<restriction>\<^sub>v"
  have dist:"((?T - ?S)\<restriction>\<^sub>v \<cdot># \<sigma>) + (?S\<restriction>\<^sub>v \<cdot># \<sigma>) = ((?T - ?S)\<restriction>\<^sub>v + ?S\<restriction>\<^sub>v) \<cdot># \<sigma>" by auto
  then have "?T\<restriction>\<^sub>v \<cdot># \<sigma> \<subseteq># ((?T - ?S)\<restriction>\<^sub>v + ?S\<restriction>\<^sub>v) \<cdot># \<sigma>" by (metis filter_diff_mset image_mset_subseteq_mono subset_eq_diff_conv subset_mset.order_refl)
  then have "\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>) \<subseteq># \<nabla># f (((?T - ?S)\<restriction>\<^sub>v + ?S\<restriction>\<^sub>v) \<cdot># \<sigma>)" by (metis image_mset_union subset_mset.le_iff_add sum_mset.union) 
  moreover have "(\<nabla># f (((?T - ?S)\<restriction>\<^sub>v + ?S\<restriction>\<^sub>v) \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2) = (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2)" using assms(5,6) by (simp add: filter_fun_def) 
  ultimately have "(\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2) \<subseteq># (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2)" by (metis filter_fun_def multiset_filter_mono) 
  then have ns_mul:"((\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2), (\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2)) \<in> ns_mul_ext ns_rel s_rel"
    using assms(4) order_pair.axioms(1) pre_order_pair_def supseteq_imp_ns_mul_ext by blast
  moreover have s_mul:"(?S\<restriction>\<^sub>p (f,2) \<cdot># \<sigma>,?T\<restriction>\<^sub>p (f,2) \<cdot># \<sigma>) \<in> s_mul_ext ?ac_rel s_rel" 
    using s_mul_ext_pres_sub_closer[OF assms(1-3)] fi by blast 
  ultimately have "(?S\<restriction>\<^sub>p (f,2) \<cdot># \<sigma> + (\<nabla># f (?S\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2), ?T\<restriction>\<^sub>p (f,2) \<cdot># \<sigma> + (\<nabla># f (?T\<restriction>\<^sub>v \<cdot># \<sigma>))\<restriction>\<^sub>p (f,2)) \<in> s_mul_ext ?ac_rel s_rel"
    using s_ns_mul_ext_union_compat[OF s_mul ns_mul] by simp
  then show ?thesis  by (simp add: aux_filter_fun_actop_subst_eq_expr)
qed

(*
    Section: Closed under substitution for ackbo, above lemmas are more general for the last case (case 3)
*)

lemma ackbo_subst:
  fixes \<sigma> :: "('f, 'v) subst"
  assumes "ackbo s t"
  shows "ackbo (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
  using assms
proof (induct s arbitrary: t rule: subterm_induct)
  case (subterm s t)
  then show ?case
  proof (cases s)
    case (Var x)
    then show ?thesis using subterm(2) by (meson ackbo_base_lhs_fun ackbo_ext_lhs_fun ackbo_eq_ackbo_wk is_VarI) 
  next
    case (Fun f fs)
    then have wt:"weight t \<le> weight (Fun f fs)" using subterm(2) ackbo_weight_eq by fastforce
    have vt:"vars_term_ms t \<subseteq># vars_term_ms (Fun f fs)" using Fun subterm(2) ackbo.simps by blast
    have vars_sub:"vars_term_ms (t \<cdot> \<sigma>) \<subseteq># vars_term_ms (s \<cdot> \<sigma>)" using vt Fun vars_term_ms_subst_mono by blast
    then consider (a) "weight (Fun f fs \<cdot> \<sigma>) > weight (t \<cdot> \<sigma>)"| (b) "weight (Fun f fs \<cdot> \<sigma>) = weight (t \<cdot> \<sigma>)" 
      using weight_stable_le[OF wt vt] dual_order.strict_iff_order by blast 
    then show ?thesis
    proof cases
      case a
      then show ?thesis using vars_sub using case_w local.Fun by blast
    next
      case b
      let ?ac_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
      have wt_eq:"weight t = weight (Fun f fs)" using b weight_stable_lt wt by (metis nat_less_le vt) 
      consider (i) "ackbo_case0 (Fun f fs) t"| (ii) "ackbo_case1 (Fun f fs) t"|
        (iii) "ackbo_case2 ackbo (Fun f fs) t"| (iv) "ackbo_case3 ackbo (Fun f fs) t"
        using ackbo.simps[of s t] subterm(2) Fun wt_eq vt by fastforce
      then show ?thesis
      proof cases
        case i
        then show ?thesis using ackbo_subterm_property Fun ackbo_case0_subterm by blast
      next
        case ii
        then obtain g gs where "t = Fun g gs" using ackbo_case1.elims(2) by blast 
        then have "ackbo_case1 (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" using Fun ackbo_case1.simps(1) subst_apply_term.simps(2) ii by auto 
        then show ?thesis using b vars_sub Fun ackbo.simps by blast
      next
        case iii
        obtain ts where term_t:"t = Fun f ts" using ackbo_case2_arg_intro iii by blast
        then have lext:"fst (lex_ext (\<lambda> x y. ((x,y) \<in> {(u,v). ackbo u v}, (x, y) \<in> ?ac_rel)) (length fs) fs ts)" using iii ackbo_case2_impl_lex_ext_s by auto 
        have local_subst_c:"\<And>u v. u \<in> set fs \<Longrightarrow> v \<in> set ts \<Longrightarrow> (u, v) \<in> {(x, y). ackbo x y} \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> {(x, y). ackbo x y}"
          using subterm(1) Fun iii by blast
        have local_subst_ac:"\<And>s t. s \<in> set fs \<Longrightarrow> t \<in> set ts \<Longrightarrow> (s, t) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" by (simp add: acstep_closed_subst) 
        have "length fs = length ts" using ackbo_case2_length_eq iii term_t by blast 
        then have "fst (lex_ext (\<lambda> x y. (ackbo x y, (x, y) \<in> ?ac_rel)) (length (map (\<lambda>t. t \<cdot> \<sigma>) fs)) (map (\<lambda>t. t \<cdot> \<sigma>) fs) (map (\<lambda>t. t \<cdot> \<sigma>) ts))"
          using list_ext_pres_sub_closer[OF lext local_subst_ac local_subst_c] iii by auto
        moreover have "length (map (\<lambda>t. t \<cdot> \<sigma>) fs) = length fs \<and> length (map (\<lambda>t. t \<cdot> \<sigma>) ts) = length ts" by simp
        ultimately have "ackbo_case2 ackbo (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" using term_t Fun ackbo_case2.simps(1)[of ackbo f "(map (\<lambda>t. t \<cdot> \<sigma>) fs)" f "(map (\<lambda>t. t \<cdot> \<sigma>) ts)"] subst_apply_term.simps(2)
          by (smt ackbo_case2_impl_not_acsym ackbo_case2_length_eq iii map_eq_conv)
        then show ?thesis using b vars_sub Fun ackbo.simps by blast
      next
        case iv
        let ?S = "actop f (Fun f fs)"
        let ?T = "actop f t"
        let ?S' = "actop f (Fun f fs \<cdot> \<sigma>)"
        let ?T' = "actop f (t \<cdot> \<sigma>)"
        let ?ackbo_rel = "{(s,t). ackbo s t}"
        let ?f = "(f,2)"
        have acsym:"f \<in> AC" using ackbo_case3.elims(2) iv by fastforce
        obtain f1 f2 where term_s:"s = Fun f [f1, f2]" using Fun iv using ackbo_case3_arg_intro by blast
        obtain t1 t2 where term_t:"t = Fun f [t1, t2]" using iv using ackbo_case3_arg_intro by blast
        have "actop f (Fun f fs) \<noteq> {#Fun f fs#} \<and> actop f t \<noteq> {#t#}" using term_s Fun[symmetric] term_t by (metis local.Fun multi_member_last trivial_Bin_facts(5)) 
        then have local_subst_c:"\<And>u v. u \<in># actop f (Fun f fs) \<Longrightarrow> v \<in># actop f t \<Longrightarrow> (u, v) \<in> {(x, y). ackbo x y} \<Longrightarrow> (u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> {(x, y). ackbo x y}"
          using subterm(1) actop_impl_subterm by (metis case_prodD case_prodI local.Fun mem_Collect_eq) 
        consider (1) "filtered_nless_rel_s ?S ?T ?f ?ac_rel ?ackbo_rel"|
          (2) "filtered_nless_rel_ns ?S ?T ?f ?ac_rel ?ackbo_rel \<and> size ?S > size ?T"|
          (3) "filtered_nless_rel_ns ?S ?T ?f ?ac_rel ?ackbo_rel \<and> size ?S = size ?T \<and> filtered_less_rel_s ?S ?T ?f ?ac_rel ?ackbo_rel"
          using iv ackbo_case3_Elim by blast
        then have "ackbo_case3_filtered_rel ?S' ?T' ?f ?ac_rel ?ackbo_rel"
        proof cases
          case 1
          show ?thesis using lemma_5_8_1[OF 1 _ local_subst_c ackbo_ord_p.order_pair_axioms acsym] ackbo_subterm_property
            by (simp add: \<open>\<And>t s. s \<rhd> t \<Longrightarrow> ackbo s t\<close> acstep_closed_subst)
        next
          case 2
          have "filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel \<or> \<not> filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel" by blast
          then show ?thesis
          proof
            assume "filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel"
            then show ?thesis by blast
          next
            assume as:"\<not> filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel"
            then have "int_size ?S - int_size ?T \<le> int_size ?S' - int_size ?T'" 
              using lemma_5_8_3[OF _ _ local_subst_c ackbo_ord_p.order_pair_axioms acsym _ as] 2
              by (simp add: ackbo_subterm_property acstep_closed_subst) 
            then have "size ?S' > size ?T'" using 2 by linarith
            then show ?thesis using lemma_5_8_2[OF _ acstep_closed_subst local_subst_c ackbo_ord_p.order_pair_axioms acsym _ as] 2 ackbo_subterm_property by blast 
          qed
        next
          case 3
          have "filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel \<or> \<not> filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel" by blast
          then show ?thesis
          proof
            assume "filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel"
            then show ?thesis by blast
          next
            assume as:"\<not> filtered_nless_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel"
            then have "int_size ?S - int_size ?T \<le> int_size ?S' - int_size ?T'" 
              using lemma_5_8_3[OF _ _ local_subst_c ackbo_ord_p.order_pair_axioms acsym _ as] 3
              by (simp add: ackbo_subterm_property acstep_closed_subst)
            then have "size ?S' \<ge> size ?T'" using 3 by linarith
            then have "size ?T' = size ?S' \<or> size ?T' < size ?S'" using le_imp_less_or_eq by blast
            then show ?thesis
            proof
              assume eq_size:"size ?T' = size ?S'"
              have fl:"filtered_nless_rel_ns ?S' ?T' ?f ?ac_rel ?ackbo_rel \<and> ((?T - ?S)\<restriction>\<^sub>v \<cdot># \<sigma>)\<restriction>\<^sub>p ?f = {#} \<and> (?T - ?S)\<restriction>\<^sub>v \<cdot># \<sigma> = \<nabla># f ((?T - ?S)\<restriction>\<^sub>v \<cdot># \<sigma>)"
                using lemma_5_8_2[OF _ acstep_closed_subst local_subst_c ackbo_ord_p.order_pair_axioms acsym _ as] ackbo_subterm_property 3 by blast
              then have "filtered_less_rel_s ?S' ?T' ?f ?ac_rel ?ackbo_rel"
                using lemma_5_8_4[OF _ acstep_closed_subst local_subst_c ackbo_ord_p.order_pair_axioms] ackbo_subterm_property fl 3 by blast
              then show ?thesis using eq_size fl by fastforce
            next
              assume "size ?T' < size ?S'"
              then show ?thesis using lemma_5_8_2[OF _ acstep_closed_subst local_subst_c ackbo_ord_p.order_pair_axioms acsym _ as] ackbo_subterm_property 3 by blast
            qed
          qed
        qed
        moreover have "Fun f fs \<cdot> \<sigma> = Fun f [f1 \<cdot> \<sigma>, f2 \<cdot> \<sigma>] \<and> (t \<cdot> \<sigma>) = Fun f [t1 \<cdot> \<sigma>, t2 \<cdot> \<sigma>]"
          using Fun term_s subst_apply_term.simps(2) term_t by (simp add: Fun)
        moreover have "length [f1 \<cdot> \<sigma>, f2 \<cdot> \<sigma>] = 2 \<and>  length [t1 \<cdot> \<sigma>, t2 \<cdot> \<sigma>] = 2" by auto
        ultimately have "ackbo_case3 ackbo (Fun f fs \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
          using acsym iv ackbo_case3.simps(1)[of ackbo f "[f1 \<cdot> \<sigma>, f2 \<cdot> \<sigma>]" f "[t1 \<cdot> \<sigma>, t2 \<cdot> \<sigma>]"] by (smt case_prodI)
        then show ?thesis using vars_sub b ackbo.simps Fun by blast
      qed
    qed
  qed
qed

lemma subseq_weight:
  assumes "subseq ss ts"
  shows "(\<Sum>s\<leftarrow>ss. weight s) \<le> (\<Sum>t\<leftarrow>ts. weight t)"
  using assms by (induct) simp+

lemma subseq_vars_term_ms:
  assumes "subseq ts ss"
  shows "vars_term_ms (Fun g ts) \<subseteq># vars_term_ms (Fun f ss)"
  using assms
proof (induct)
  case (list_emb_Cons xs ys y)
  obtain m1 m2 where id: "\<Union>#image_mset vars_term_ms (mset xs) = m1" 
    "\<Union>#image_mset vars_term_ms (mset ys) = m2" by auto
  from list_emb_Cons
  have IH: "m1 \<subseteq># m2" by (auto simp: id)
  also have "\<dots> \<subseteq># vars_term_ms y + m2" by (rule mset_subset_eq_add_right)
  finally show ?case by (simp add: id)
qed auto

lemma wqo_on_map':
  fixes P and Q and h
  defines "P' \<equiv> \<lambda>x y. P x y \<and> Q (h x) (h y)"
  assumes "wqo_on P\<^sup>=\<^sup>= A"
    and "wqo_on Q B"
    and subset: "h ` A \<subseteq> B"
  shows "wqo_on P'\<^sup>=\<^sup>= A"
proof
  let ?Q = "\<lambda>x y. Q (h x) (h y)"
  from \<open>wqo_on P\<^sup>=\<^sup>= A\<close> have "transp_on P\<^sup>=\<^sup>= A"
    by (rule wqo_on_imp_transp_on)
  then show "transp_on P'\<^sup>=\<^sup>= A"
    using \<open>wqo_on Q B\<close> and subset
    unfolding wqo_on_def transp_on_def P'_def by blast

  from \<open>wqo_on P\<^sup>=\<^sup>= A\<close> have "almost_full_on P\<^sup>=\<^sup>= A"
    by (rule wqo_on_imp_almost_full_on)
  from \<open>wqo_on Q B\<close> have "almost_full_on Q B"
    by (rule wqo_on_imp_almost_full_on)

  show "almost_full_on P'\<^sup>=\<^sup>= A"
  proof
    fix f
    assume *: "\<forall>i::nat. f i \<in> A"
    from almost_full_on_imp_homogeneous_subseq [OF \<open>almost_full_on P\<^sup>=\<^sup>= A\<close> this]
    obtain g :: "nat \<Rightarrow> nat"
      where g: "\<And>i j. i < j \<Longrightarrow> g i < g j"
        and **: "\<forall>i. f (g i) \<in> A \<and> P\<^sup>=\<^sup>= (f (g i)) (f (g (Suc i)))"
      using * by auto
    from chain_transp_on_less [OF ** \<open>transp_on P\<^sup>=\<^sup>= A\<close>]
    have **: "\<And>i j. i < j \<Longrightarrow> P\<^sup>=\<^sup>= (f (g i)) (f (g j))" .
    let ?g = "\<lambda>i. h (f (g i))"
    from * and subset have B: "\<And>i. ?g i \<in> B" by auto
    with \<open>almost_full_on Q B\<close> [unfolded almost_full_on_def good_def, THEN bspec, of ?g]
    obtain i j :: nat
      where "i < j" and "Q (?g i) (?g j)" by blast
    with ** [OF \<open>i < j\<close>] have "P'\<^sup>=\<^sup>= (f (g i)) (f (g j))"
      by (auto simp: P'_def)
    with g [OF \<open>i < j\<close>] show "good P'\<^sup>=\<^sup>= f" by (auto simp: good_def)
  qed
qed

lemma ackbo_simplification_order:
  assumes "wqo_on pr_strict\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV"
  shows "simplification_order {(x :: ('f, 'v) term, y). ackbo x y}" (is "simplification_order ?ackbo")
proof (unfold simplification_order_def)
  let ?P = "\<lambda>x y. pr_strict x y \<and> w x \<ge> w y"
  have "range w \<subseteq> UNIV" by auto
  from wqo_on_map' [OF assms wqo_on_UNIV_nat this]
  have "wqo_on ?P\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV" by (auto simp: conversep_iff [abs_def])
  moreover have "rewrite_order ?ackbo"
  proof -
    have "rewrite_relation ?ackbo" unfolding rewrite_relation_def using ackbo_subst ackbo_ctxt by blast
    moreover have "strict_order ?ackbo" using ackbo_irr ackbo_ord_p.trans_S unfolding irrefl_def by blast
    ultimately show ?thesis unfolding rewrite_order_def by simp
  qed
  moreover have "emb UNIV ?P \<subseteq> ?ackbo"
  proof (rule emb_subsetI)
    show "\<And>s t u. (s, t) \<in> ?ackbo \<Longrightarrow> (t, u) \<in> ?ackbo \<Longrightarrow> (s, u) \<in> ?ackbo" using ackbo_ord_p.trans_S unfolding trans_def by blast
    show "\<And>C s t. (s, t) \<in> ?ackbo \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> ?ackbo" using ackbo_ctxt by auto
    show "\<And>\<sigma> s t. (s, t) \<in> ?ackbo \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?ackbo" using ackbo_subst by blast
    show "\<And>t f ts. t \<in> set ts \<Longrightarrow> (Fun f ts, t) \<in> ?ackbo" using ackbo_subterm_property by blast
    show "\<And>f g ss ts. pr_strict (f, length ss) (g, length ts) \<and> w (g, length ts) \<le> w (f, length ss) \<Longrightarrow> subseq ts ss \<Longrightarrow> (Fun f ss, Fun g ts) \<in> ?ackbo"
    proof -
      fix f g and ss ts :: "('f, 'v) term list"
      assume strict: "pr_strict (f, length ss) (g, length ts) \<and> w (f, length ss) \<ge> w (g, length ts)"
      assume subseq:"subseq ts ss"
      have vars:"vars_term_ms (Fun g ts) \<subseteq># vars_term_ms (Fun f ss)" using subseq_vars_term_ms[OF subseq] by auto
      consider (a) "weight (Fun g ts) < weight (Fun f ss)"| (b) "weight (Fun g ts) = weight (Fun f ss)"
        using subseq_weight[OF subseq] strict by fastforce
      then show "(Fun f ss, Fun g ts) \<in> ?ackbo"
      proof cases
        case a
        then show ?thesis using vars ackbo.simps by fastforce
      next
        case b
        then have "ackbo_case1 (Fun f ss) (Fun g ts)" using strict by simp
        then show ?thesis using vars ackbo.simps b by fastforce
      qed
    qed
  qed
  ultimately show "rewrite_order ?ackbo \<and> (\<exists>Q. wqo_on Q\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV \<and> emb UNIV Q \<subseteq> ?ackbo)" using assms by blast
qed

lemma subterm_ackbo_simplification_order:
  shows "subterm_simplification_order {(s,t). ackbo s t}"
  unfolding subterm_simplification_order_def rewrite_order_def rewrite_relation_def
  by (metis ackbo_subst ackbo_ctxt ackbo_ord_p.trans_S ackbo_subterm_property ackbo_irr case_prodD case_prodI irrefl_def mem_Collect_eq subrelI) 

abbreviation ackbo_S where "ackbo_S \<equiv> {(s,t). ackbo s t}"
abbreviation ackbo_NS where "ackbo_NS \<equiv> ackbo_S\<^sup>* \<union> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"

lemma varcondS_ac:
  shows "\<And> l r. (l,r) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow> vars_term r \<subseteq> vars_term l"
proof
  fix l :: "('f, 'a) term"
  fix r x
  assume "(l, r) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
  from ac_terms_share_vars[OF this] have "vars_term_ms l = vars_term_ms r" .
  then show "x \<in> vars_term r \<Longrightarrow> x \<in> vars_term l"
    using set_mset_vars_term_ms by fastforce 
qed

lemma varcondS_ackbo:
  shows "ackbo l r \<Longrightarrow> vars_term r \<subseteq> vars_term l"
proof
  fix l :: "('f, 'a) term"
  fix r x
  assume "ackbo l r"
  then have "vars_term_ms r \<subseteq># vars_term_ms l" using ackbo.simps by blast 
  then show "x \<in> vars_term r \<Longrightarrow> x \<in> vars_term l" using mset_subsetD set_mset_mono by fastforce 
qed

lemma ackbo_orientation_imp_SN:
  fixes S :: "('f, 'v) trs"
  assumes "finite F"
    and "funas_trs (S \<union> AC_trs AC AC) \<subseteq> F"
    and "S \<subseteq> ackbo_S"
    and "S \<union> (AC_trs AC AC) \<subseteq> ackbo_NS" (is "?S \<union> ?AC \<subseteq> _")
  shows "SN_rel (rstep S) (rstep (S \<union> AC_trs AC AC))"
proof -
  have vars: "\<And> l r. (l, r) \<in> (?S \<union> ?AC) \<Longrightarrow> vars_term r \<subseteq> vars_term l" using varcondS_ackbo varcondS_ac assms(4)
    by (metis (no_types, lifting) Un_iff assms(3) case_prodD conversion_def mem_Collect_eq r_into_rtrancl rstep_rule subsetCE symcl_acstep)
  have  "rewrite_relation (ackbo_S)" unfolding rewrite_relation_def by (simp add: ackbo_subst ackbo_ctxt)
  then have "rewrite_relation (ackbo_S\<^sup>*)" unfolding rewrite_relation_def using ackbo_ord_p.trancl_S rtrancl_eq_or_trancl
  proof -
    have "\<forall>t ta r. ((t::('f, 'a) Term.term, ta) \<in> r\<^sup>*) = (t = ta \<or> t \<noteq> ta \<and> (t, ta) \<in> r\<^sup>+)"
      using rtrancl_eq_or_trancl by blast
    then show "(\<forall>t ta c. (t::('f, 'a) Term.term, ta) \<in> ackbo_S\<^sup>* \<longrightarrow> (c\<langle>t\<rangle>, c\<langle>ta\<rangle>) \<in> ackbo_S\<^sup>*) \<and> (\<forall>t ta f. (t::('f, 'a) Term.term, ta) \<in> ackbo_S\<^sup>* \<longrightarrow> (t \<cdot> f::('f, 'a) Term.term, ta \<cdot> f) \<in> ackbo_S\<^sup>*)"
      by (metis (no_types) \<open>rewrite_relation ackbo_S\<close> ackbo_ord_p.trancl_S rewrite_relation_def)
  qed
  moreover have "rewrite_relation ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" unfolding rewrite_relation_def
    by (metis acstep_closed_subst conversion_def rsteps_closed_ctxt symcl_acstep) 
  ultimately have rew: "rewrite_relation ackbo_NS" unfolding rewrite_relation_def using rewrite_relation_def by auto
  have "ackbo_NS\<^sup>* O ackbo_S O ackbo_NS\<^sup>* \<subseteq> ackbo_S"
    by (smt ackbo_ord_p.NS_O_S(1) ackbo_ord_p.S_O_rtrancl_NS(1) ackbo_ord_p.rtrancl_NS ackbo_ord_p.trancl_S relto_pair.NS_O_S(2) relto_pair.compat_S_NS rtrancl_Un_rtrancl)
  from subterm_simplification_order_on_equations_imp_SN[OF subterm_ackbo_simplification_order rew vars assms(1) _ _ _ _ this]
  show ?thesis
    using assms(2-4)
    by fastforce
qed

lemma name_me_redpair:
  fixes S :: "('f, 'v) trs"
  assumes "finite F"
    and "funas_trs (S \<union> AC_trs AC AC) \<subseteq> F"
    and "S \<subseteq> ackbo_S"
    and "S \<union> (AC_trs AC AC) \<subseteq> ackbo_NS" (is "?NS \<subseteq> _")
  shows "redtriple_order ((relto (rstep S) (rstep (AC_trs AC AC)))\<^sup>+) ((rstep (S \<union> AC_trs AC AC))\<^sup>*) ((rstep (S \<union> AC_trs AC AC))\<^sup>*)"
    (is "redtriple_order ?S ?NS _")
proof (unfold_locales)
  have sn_rel: "SN_rel (rstep S) (rstep (S \<union> AC_trs AC AC))" using ackbo_orientation_imp_SN[OF assms] by auto
  show "SN ?S" using sn_rel by (simp add: SN_rel_imp_SN_relto relto_trancl_conv rstep_simps)
  show "ctxt.closed ?NS" by (force simp add: ctxt_closed_rsteps)
  show "subst.closed ?S" by (simp add: subst.closed_relto subst.closed_trancl subst_closed_rstep)
  show "subst.closed ?NS" by (force simp add: subst_closed_rsteps)
  show "?NS O ?S \<subseteq> ?S" by (simp add: relto_pair.compat_NS_S rstep_simps)
  show "?S O ?NS \<subseteq> ?S" by (simp add: relto_pair.compat_S_NS rstep_simps)
  show "?S \<subseteq> ?NS" by (metis le_sup_iff r_into_rtrancl relto_trancl_conv relto_trancl_subset rstep_union rtrancl_idemp rtrancl_idemp_self_comp subsetI trancl_unfold_left) 
  show "refl ?NS" using refl_rtrancl by blast
  show "trans ?S" using trans_trancl by blast
  show "trans ?NS" by (simp add: trans_rtrancl)
qed

end

end
